/******************************************************************************
 * A x e l                                                                    *
 ******************************************************************************
 * QImplicitCurvesArrangement                                                 *
 * 2006-11-16                                                                 *
 * Julien Wintz                                                               *
 ******************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis                    *
 ******************************************************************************
 * TODO : intersection are computed twice:                                    *
 * - once in regularity test                                                  *
 * - once in regions computation                                              *
 *                                                                            *
 * FIXED : intersections as well as singular points are always computed in the*
 * regularity test because the latter is performed for ALL cells whereas the  *
 * regions computations is ONLY performed for the bottommost cells of the     *
 * octree.                                                                    *
 *                                                                            *
 * TODO : intersection points and singular points are ALWAYS computed but not *
 * necessarily taken into accounts, if they have already been computed. It    *
 * should be computed only if necessary instead of systematically.            *
 * Sadly, it depends on the exactness of the solver used. (Currently, Synaps  *
 * Bezier Isolation Solver).                                                  *
 ******************************************************************************/

# include <stdio.h>

# include <QData/QCell.h>

# include <QGui/color.h>

# include <QAxel/PluginManager.h>

# include <QSynaps/QSynapsInterface.h>

# include <QArrangement/Curves/QImplicitCurvesArrangement.h>

using namespace std ;
using namespace QAxel ;
using namespace QData ;

namespace QArrangement {

QImplicitCurvesArrangement::QImplicitCurvesArrangement(MainWindow * mainwindow, 
						       double islPrecision, 
						       double sbdPrecision, 
						       double mulPrecision, 
						       double cellPrecision, 
						       double cellPrecisionMax) : 
    QArrangement(QString("ImplicitCurve2D"), cellPrecision, cellPrecisionMax),
    s_islPrecision(islPrecision),
    s_sbdPrecision(sbdPrecision),
    s_mulPrecision(mulPrecision)
{
    this->mainwindow = mainwindow ;
}

bool QImplicitCurvesArrangement::regular(Object_t * o, Node_t * n)
{	
    QImplicitCurve2D * c = dynamic_cast<QImplicitCurve2D *>(o) ;
    QSynapsInterface * iSynaps = dynamic_cast<QSynapsInterface *>(PluginManager::instance()->plugin("Synaps")) ;
    Cell_t * bb = n->GetCell() ;
    
    // cout << "QImplicitCurvesArrangement::regular() for " << BG_GREEN << bb->name().toAscii().constData() << " " << n->depth << NOCOLOR << flush ;
    
    /****************************************************************************
     * Without Dynamic Programming
     * 
     * Intersections with the curve are systematically computed for each cell
     * regardless the fact some of them (the outer ones actually) have already
     * been computed previously at an upper level of the subdivision, for its
     * parent cell.
     * 
     * In this case, each arrangement vertex defined by the intersection of the
     * curve with the cell will generate as much logical point as the depth of 
     * the subdivision in the worse case.
     * 
     * Each of this point is, when computing an arrangement of algebraic curves,
     * computed with a isolation bezier solver which provides approximate 
     * solutions.
     * It might therefor happen, that an intersection point adjacent to two 
     * regions in two different cells may have coordinates which differ from 
     * a cell to another.
     * 
     * This can be solved by applying dynamic programming, where, an intersection
     * of the curve with a cell is computed only once, and then, propagated from
     * a mother to a child, and from a child to its sisters.
     ****************************************************************************/

    // QList<Point_t *> intersectionPoints = iSynaps->intersectionPoints(c, bb) ;
    // QList<Point_t *> singularPoints = iSynaps->singularPoints(c, bb) ;
        
    /****************************************************************************
     * With Dynamic Programming
     * 
     * Cells already have inherited intersection from their parent.
     * Here we also check already computed intersections for their 
     * sisters (or neighboors).
     * 
     * Because the solver used to get roots is approximate, we have to get all
     * roots (roots on each side) at once, and then, dispatch them on the borders.
     * 
     * For each border (North|South|East|West), we perform inheritance of 
     * previously computed roots:
     * - if they are not already present, which means, if they are, they have been
     * already inherited from the parent cell.
     * - if they have not been computed in a neighbor.
     * - else, we compute them.
     ****************************************************************************/
    
    QList<Point_t *> intersectionPoints ;
    QList<Point_t *> singularPoints ;
    QList<Point_t *> xtremPoints ; 
    QList<Point_t *> ytremPoints ;
    QList<Point_t *> rootPoints = iSynaps->intersectionPoints(c, bb, s_islPrecision) ;
    QList<Point_t *> westIntersections  ; foreach(QPoint3f * p, rootPoints) if(p->x() == bb->xmin()) westIntersections  << p ;
    QList<Point_t *> eastIntersections  ; foreach(QPoint3f * p, rootPoints) if(p->x() == bb->xmax()) eastIntersections  << p ;
    QList<Point_t *> southIntersections ; foreach(QPoint3f * p, rootPoints) if(p->y() == bb->ymin() && !westIntersections.contains(p) && !eastIntersections.contains(p)) southIntersections << p ;
    QList<Point_t *> northIntersections ; foreach(QPoint3f * p, rootPoints) if(p->y() == bb->ymax() && !westIntersections.contains(p) && !eastIntersections.contains(p)) northIntersections << p ;
    
    if(bb->westIntersections().size() == 0)
	if(n->getWestNeighborNode() && n->getWestNeighborNode()->depth == n->depth && n->getWestNeighborNode()->GetCell()->eastIntersections().size() > 0) {
	    // cout << FG_YELLOW << bb->name().toAscii().constData() << ": West intersections inherited from neighbor" << NOCOLOR << endl ;
	    bb->setWestIntersections(n->getWestNeighborNode()->GetCell()->eastIntersections()) ;
	} else {
	    // cout << FG_MAGENTA << bb->name().toAscii().constData() << ": West intersections were not yet computed" << NOCOLOR << endl ;
	    bb->setWestIntersections(westIntersections) ;
	    // foreach(Point_t * p, westIntersections) { p->setSize(0.005) ; mainwindow->viewer()->addObject(p) ; }
	}
    // else cout << FG_CYAN << bb->name().toAscii().constData() << ": West intersections wre already computed" << NOCOLOR << endl ;
    intersectionPoints << bb->westIntersections() ;
    
    if(bb->eastIntersections().size() == 0)
	if(n->getEastNeighborNode() && n->getEastNeighborNode()->depth == n->depth && n->getEastNeighborNode()->GetCell()->westIntersections().size() > 0) {
	    // cout << FG_YELLOW << bb->name().toAscii().constData() << ": East intersections inherited from neighbor" << NOCOLOR << endl ;
	    bb->setEastIntersections(n->getEastNeighborNode()->GetCell()->westIntersections()) ;
	} else {
	    // cout << FG_MAGENTA << bb->name().toAscii().constData() << ": East intersections were not yet computed" << NOCOLOR << endl ;
	    bb->setEastIntersections(eastIntersections) ;
	    // foreach(Point_t * p, eastIntersections) { p->setSize(0.005) ; mainwindow->viewer()->addObject(p) ; }
	}
    // else cout << FG_CYAN << bb->name().toAscii().constData() << ": East intersections wre already computed" << NOCOLOR << endl ;
    intersectionPoints << bb->eastIntersections() ;
    
    if(bb->southIntersections().size() == 0)
	if(n->getSouthNeighborNode() && n->getSouthNeighborNode()->depth == n->depth && n->getSouthNeighborNode()->GetCell()->northIntersections().size() > 0) {
	    // cout << FG_YELLOW << bb->name().toAscii().constData() << ": South intersections inherited from neighbor" << NOCOLOR << endl ;
	    bb->setSouthIntersections(n->getSouthNeighborNode()->GetCell()->northIntersections()) ;
	} else {
	    // cout << FG_MAGENTA << bb->name().toAscii().constData() << ": South intersections were not yet computed" << NOCOLOR << endl ;
	    bb->setSouthIntersections(southIntersections) ;
	    // foreach(Point_t * p, southIntersections) { p->setSize(0.005) ; mainwindow->viewer()->addObject(p) ; }
	}
    // else cout << FG_CYAN << bb->name().toAscii().constData() << ": South intersections wre already computed" << NOCOLOR << endl ;
    intersectionPoints << bb->southIntersections() ;
    
    if(bb->northIntersections().size() == 0)
	if(n->getNorthNeighborNode() && n->getNorthNeighborNode()->depth == n->depth && n->getNorthNeighborNode()->GetCell()->southIntersections().size() > 0) {
	    // cout << FG_YELLOW << bb->name().toAscii().constData() << ": North intersections inherited from neighbor" << NOCOLOR << endl ;
	    bb->setNorthIntersections(n->getNorthNeighborNode()->GetCell()->southIntersections()) ;
	} else {
	    // cout << FG_MAGENTA << bb->name().toAscii().constData() << ": North intersections were already computed" << NOCOLOR << endl ;
	    bb->setNorthIntersections(northIntersections) ;
	    // foreach(Point_t * p, northIntersections) { p->setSize(0.005) ; mainwindow->viewer()->addObject(p) ; }
	}
    // else cout << FG_CYAN << bb->name().toAscii().constData() << ": North intersections wre already computed" << NOCOLOR << endl ;
    intersectionPoints << bb->northIntersections() ;
    
    if(bb->singularPoints().size() == 0)
	bb->setSingularPoints(iSynaps->singularPoints(c, bb/*, s_sbdPrecision*/)) ;
    singularPoints << bb->singularPoints() ;
	
    if(bb->xtremPoints().size() == 0)
	bb->setXtremPoints(iSynaps->xtremPoints(c,bb,s_sbdPrecision)) ;
    xtremPoints << bb->xtremPoints() ;

    if(bb->ytremPoints().size() == 0)
	bb->setYtremPoints(iSynaps->ytremPoints(c,bb,s_sbdPrecision)) ;
    ytremPoints << bb->ytremPoints() ;
    
    /****************************************************************************
     * End of dynamic programming
     ****************************************************************************/
    
    /****************************************************************************
     * Actual regularity test
     ****************************************************************************/
    
    // cout << __func__ << ": cell " << bb->name().toAscii().constData() << " : " << intersectionPoints.size() << " intersection points" << endl ;
    // cout << __func__ << ": cell " << bb->name().toAscii().constData() << " : " << singularPoints.size() << " singular points" << endl ;
    // cout << __func__ << ": cell " << bb->name().toAscii().constData() << " : " << xtremPoints.size() << " xtrem points" << endl ;
    // cout << __func__ << ": cell " << bb->name().toAscii().constData() << " : " << ytremPoints.size() << " ytrem points" << endl ;
	
    if(xtremPoints.size() + ytremPoints.size() <= 1 && intersectionPoints.size() <= 2)
	return true ;
	
    // if(xtremPoints.size() + ytremPoints.size() <= 1 && intersectionPoints.size() == 4)
    // 	return true ;

    //********************************************************
    // Removing points of multiplicity > 1
    //********************************************************
    if(singularPoints.size()) {
	int i = 0 ;
	while(i < xtremPoints.size())
	    foreach(Point_t * p, singularPoints)
		if(xtremPoints[i]->equals(p, s_mulPrecision)) { 
		    // cout << FG_YELLOW << "removing a xtrem point!!!" << NOCOLOR << endl ; 
		    xtremPoints.removeAt(i) ; 
		}
	// if(iSynaps->isOn(xtremPoints[i], iSynaps->computeHessianCurve(c))) xtremPoints.removeAt(i) ;
		else i++ ;
			
	int j = 0 ;
	while(j < ytremPoints.size())
	    foreach(Point_t * p, singularPoints)
		if(ytremPoints[j]->equals(p, s_mulPrecision)) { 
		    // cout << FG_YELLOW << "removing a ytrem point!!!" << NOCOLOR << endl ; 
		    ytremPoints.removeAt(j) ; 
		}
	// if(iSynaps->isOn(ytremPoints[i], iSynaps->computeHessianCurve(c))) ytremPoints.removeAt(i) ;
		else j++ ;
    }
	
    //********************************************************
    // TODO:
    // We assume the curves to be squrefree. We can therefore
    // remove multiple critical points by checking if they
    // lie in the second derivative curve.
    //********************************************************
	
    // cout << "Checking for singularities: # of singularities: " << singularPoints.size() << " number of branches for the first: "
    // 	 << iSynaps->numberOfBranches(singularPoints.first(), c, bb) << " #xtrem : " << xtremPoints.size() << " #ytrem : " << ytremPoints.size() << endl ;
    // getchar();
	
    if(xtremPoints.size() + ytremPoints.size() == 0 && singularPoints.size() == 1 && intersectionPoints.size() == iSynaps->numberOfBranches(singularPoints.first(), c, bb))
	return true ;

    return false ;
}

/*!*************************************************************************************
  A cell can be subdivided following different policies (cf. QArrangement generic class)
  This function chooses a "center" for the subdivision and then calls the subdivision
  of the cell itself.
  The result of the subdivision is always given in the same order: TL, TR, BL, BR
 
  Whatever the policy, a cell is divided in 4 or 8 parts. It can therefor be embedded in
  a quadtree or an octree.
 ***************************************************************************************/ 
QList<Cell_t *> QImplicitCurvesArrangement::subdivide(Object_t * o, Node_t * n)
{
    Cell_t * cell = n->GetCell() ;
    QList<Point_t *> singularPoints = cell->singularPoints() ;
    QList<Point_t *> xtremPoints    = cell->xtremPoints() ;
    QList<Point_t *> ytremPoints    = cell->ytremPoints() ;
    QList<Cell_t *> l ;  
  
    if(a_policy == Center)
	l << cell->subdivide2D(cell->center()) ;

    Q_UNUSED(o) ;
    Q_UNUSED(singularPoints) ;
    Q_UNUSED(xtremPoints) ;
    Q_UNUSED(ytremPoints) ;
  
    return l ;
}

QList<Region_t *> QImplicitCurvesArrangement::topology(Object_t * o, Node_t * n)
{
    Cell_t * bb = n->GetCell() ;
	
    // cout << __func__ << " for " << BG_RED << bb->name().toAscii().constData() << NOCOLOR << endl ;
    QImplicitCurve2D * c = dynamic_cast<QImplicitCurve2D *>(o) ;
  
    QList<Region_t *> l ;
  
    Vertex_t * se = new Vertex_t(bb->xmax(), bb->ymin()) ; se->setName("se") ;
    Vertex_t * sw = new Vertex_t(bb->xmin(), bb->ymin()) ; sw->setName("sw") ;
    Vertex_t * ne = new Vertex_t(bb->xmax(), bb->ymax()) ; ne->setName("ne") ;
    Vertex_t * nw = new Vertex_t(bb->xmin(), bb->ymax()) ; nw->setName("nw") ;

    QList<Point_t *> intersectionPoints ;
    intersectionPoints << bb->northIntersections() ;
    intersectionPoints << bb->eastIntersections() ;
    intersectionPoints << bb->southIntersections() ;
    intersectionPoints << bb->westIntersections() ;

    QList<Point_t *> singularPoints = bb->singularPoints() ;
    QList<Point_t *> xtremPoints = bb->xtremPoints() ;
    QList<Point_t *> ytremPoints = bb->ytremPoints() ;

    if(singularPoints.size() == 0) {
    
    	if(intersectionPoints.size() == 0) {
	    
	    //*************************************************************************************
	    // If there is no intersection points with the border of the cell, we create a region
	    // representing the whole cell.
	    //*************************************************************************************
      
	    // An empty cell
	    Region_t * f = new Region_t() ;
	    f->addVertices(QList<Vertex_t *>() << nw << sw << se << ne) ;
	    f->addEdge(nw, sw, bb) ;
	    f->addEdge(sw, se, bb) ;
	    f->addEdge(se, ne, bb) ;
	    f->addEdge(ne, nw, bb) ;
	    l << f ;
	}
    
	else if (intersectionPoints.size() == 2) {
      
	    //*************************************************************************************
	    // If there are two intersection points with the border of the cell, we create two
	    // regions according to how these intersections points are diposed.
	    //
	    // 3 cases:
	    // - There is no (x|y)trem point.
	    // - There is at most 1 xtrem point.
	    // - There is at most 1 ytrem point.
	    //*************************************************************************************
      
	    QList<Vertex_t *> rvertices = clockwiseSubdivision(intersectionPoints, bb, nw, ne, se, sw) ;
	    // cout << "Cell:" << endl ;
	    // cout << bb << endl ;
	    // cout << "Clockwise subdivision:" << endl ;
	    // foreach(Vertex_t * v, rvertices) cout << v << endl ;

	    while(rvertices.size()>1) {
		// cout << "---- Computing a region ----" << endl ;
		int n = 0 ;
		while(isCellBorder(rvertices[n+1], bb)) n++ ;
		Region_t * f = new Region_t ;
		for(int i = 0 ; i < n+2 ; i++) {
		    f->addVertex(rvertices[i]) ;
		    // cout << "pushing vertex " << rvertices[i]->name().toAscii().constData() << endl ;
		}
		for(int i = 0 ; i < n+1 ; i++) {
		    f->addEdge(rvertices[i+1], rvertices[i], bb) ;
		    // cout << "pushing edge [" << rvertices[i]->name().toAscii().constData() << ", " 
		    //      << rvertices[i+1]->name().toAscii().constData() << ", bb]" << endl ;
		}

		//********************************************************************************
		// There is no (x|y)trem point.
		//********************************************************************************
		if(xtremPoints.size() + ytremPoints.size() == 0)
		    f->addEdge(rvertices[0], rvertices[n+1], c) ;

		//********************************************************************************
		// There is at most 1 xtrem point.
		//********************************************************************************
		if(xtremPoints.size() == 1) {
		    Vertex_t * xtremPoint = new Vertex_t(xtremPoints[0]) ;
		    f->addVertex(xtremPoint) ;
		    f->addEdge(rvertices[0],   xtremPoint, c) ;
		    f->addEdge(xtremPoint, rvertices[n+1], c) ;
		}

		//********************************************************************************
		// There is at most 1 ytrem point.
		//********************************************************************************
		if(ytremPoints.size() == 1) {
		    Vertex_t * ytremPoint = new Vertex_t(ytremPoints[0]) ;
		    f->addVertex(ytremPoint) ;
		    f->addEdge(rvertices[0],   ytremPoint, c) ;
		    f->addEdge(ytremPoint, rvertices[n+1], c) ;
		}
		
		// cout << "pushing edge [" << rvertices[0]->name().toAscii().constData() << ", " << rvertices[n+1]->name().toAscii().constData() << "]" << endl ;
		for(int i = 0 ; i < n+1 ; i++) rvertices.takeFirst() ;
		l << f ;
	    }
      
	} else qFatal("Should never get here! No singularity, Other situation than 0 or 2 intersection points with border zhile getting topology of regions") ;

    } else if (singularPoints.size() == 1) {
    
	//**************************************************************************************
	// There is only one singularity in the cell (thanks to the regularity criteria).
	// We build a star-shaped structure with it and with the intersections on the border of
	// the cell.
	//
	// First, we build a list of points on each side of the border.
	// Second, we order these list in clockwise order 
	//
	//**************************************************************************************
		
    	// Vertex_t * s = dynamic_cast<Vertex_t *>(singularPoints[0]) ;
	Vertex_t * s = new Vertex_t(singularPoints[0]) ;

    	QList<Vertex_t *> rvertices = clockwiseSubdivision(intersectionPoints, bb, nw, ne, se, sw) ;
		
    	while(rvertices.size()>1) {
	    // cout << "---- Computing a region ---- " << endl ;
	    int n = 0 ;
	    while(isCellBorder(rvertices[n+1], bb)) n++ ;
	    Region_t * f = new Region_t ;
	    for(int i = 0 ; i < n+2 ; i++) {
		f->addVertex(rvertices[i]) ;
		// cout << "pushing vertex " << rvertices[i]->name().toAscii().constData() << endl ;
	    }
	    for(int i = 0 ; i < n+1 ; i++) {
		f->addEdge(rvertices[i+1], rvertices[i], bb) ;
		// cout << "pushing edge [" << rvertices[i]->name().toAscii().constData() << ", " 
		// << rvertices[i+1]->name().toAscii().constData() << ", bb]" << endl ;
	    }
	    f->addVertex(s) ;
	    f->addEdge(rvertices[0], s, c) ;
	    // cout << "pushing edge [s, " << rvertices[0]->name().toAscii().constData() << ", c]" << endl ;
	    //f->addEdge(s, rvertices[n+1], c) ;
	    f->addEdge(s, rvertices[n+1], c) ;
	    // cout << "pushing edge [s, " << rvertices[n+1]->name().toAscii().constData() << ", c]" << endl ;
	    for(int i = 0 ; i < n+1 ; i++) rvertices.takeFirst() ;
	    l << f ;
	}
    } else {
	qFatal("Should never get here! With Singularity");
    }

    return l ;	
}

} ; // namespace QArrangement ;
