/*****************************************************************************
 * A x e l
 *****************************************************************************
 * QMesh
 * 2006-01-25
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :
 ****************************************************************************/

# include <QColor>
# include <QAxel/QAxelObject.h>

namespace QAxel {

QAxelObject::QAxelObject(QAxelObject * parent)
{
    this->m_parent = parent ;
    this->blacklisted_ = false ;
    this->hidden_ = false ;
    this->selected_ = false ;
    this->mode_ = None ;
    this->frame_ = new ManipulatedFrame() ;
    
    if(m_parent != 0)
	m_parent->addChild(this) ;
}

QAxelObject::~QAxelObject(void) 
{
    delete color_ ;
    delete frame_ ;
}

/***************************************************************************** 
 * Initialize static members
 ****************************************************************************/
QColor * QAxelObject::selectionColor_ = new QColor(255, 190, 255, 128) ;
QColor * QAxelObject::grabbingColor_ = new QColor(255, 255, 0) ;

void QAxelObject::select(void) 
{ 
    selected_ = true ; 
}

void QAxelObject::deselect(void) 
{
    selected_ = false ; 
}

bool QAxelObject::selected(void) 
{ 
    return selected_ ; 
}

ManipulatedFrame * QAxelObject::frame(void) 
{ 
    return frame_ ; 
}

QString QAxelObject::position(void) 
{
    QString x, y, z ;
    return QString(
	"(" + x.setNum(frame_->position().x) +
	", "+ y.setNum(frame_->position().y) +
	", "+ z.setNum(frame_->position().z) + ")"
        ) ;
}
    
QString QAxelObject::orientation(void) 
{
    QString x, y, z, angle ;
    return QString(
	"(" + x.setNum(frame_->orientation().axis().x) +
	", "+ y.setNum(frame_->orientation().axis().y) +
	", "+ z.setNum(frame_->orientation().axis().z) +
	", "+ angle.setNum(frame_->orientation().angle()) + ")"
        ) ;
}

QColor * QAxelObject::color(void) 
{
    return color_ ;
}

QColor * QAxelObject::selectionColor(void) 
{
    return selectionColor_ ;
}

QColor * QAxelObject::grabbingColor(void) 
{
    return grabbingColor_ ;
}

void QAxelObject::setColor(QColor color)
{
    color_ = new QColor(color) ;
}

void QAxelObject::setObjectColor(QColor * color) 
{
    color_ = color ;
}

void QAxelObject::setRGB(int r, int g, int b) 
{
    color_->setRed(r) ;
    color_->setGreen(g) ;
    color_->setBlue(b) ;
}

void QAxelObject::setRGBA(int r, int g, int b, int a) 
{
    color_->setRed(r) ;
    color_->setGreen(g) ;
    color_->setBlue(b) ;
    color_->setAlpha(a) ;
}

void QAxelObject::setSelectionColor(QColor * color) 
{
    selectionColor_ = color ;
}

void QAxelObject::setGrabbingColor(QColor * color) 
{
    grabbingColor_ = color ;
}

void QAxelObject::glColor(void) 
{
    if (frame_->grabsMouse()) {
	if (frame_->isManipulated()) ; 	
	else glGrabbingColor() ;
    } else {
	if(selected()) glSelectionColor() ;
	else glObjectColor() ;
    }
}

void QAxelObject::glColorAlpha(void) 
{
    if (frame_->grabsMouse()) {
	if (frame_->isManipulated()) ; 	
	else glGrabbingColor() ;
    } else {
	if(selected()) glSelectionColorAlpha() ;
	else glObjectColorAlpha() ;
    }
}

void QAxelObject::glObjectColor(void) 
{
    glColor4ub(color_->red(), color_->green(), color_->blue(), color_->alpha()) ;
}

void QAxelObject::glObjectColorAlpha(void) 
{
    glColor4ub(color_->red(), color_->green(), color_->blue(), color_->alpha()) ;
}

void QAxelObject::glSelectionColor(void) 
{
    glColor4ub(selectionColor_->red(), selectionColor_->green(), selectionColor_->blue(), color_->alpha()) ;
}

void QAxelObject::glSelectionColorAlpha(void) 
{
    glColor4ub(selectionColor_->red(), selectionColor_->green(), selectionColor_->blue(), selectionColor_->alpha()) ;
}

void QAxelObject::glGrabbingColor(void) 
{
    glColor4ub(grabbingColor_->red(), grabbingColor_->green(), grabbingColor_->blue(), grabbingColor_->alpha()) ;
}

/****************************************************************************/
/*! Returns the \e parent for the QAxelObject. This parent is the one
    under which the QAxelObject is. It may be \c 0 if the QAxelObject
    has no parent yet, or is a \e toplevel QAxelObject.
*/
/****************************************************************************/
QAxelObject * QAxelObject::parent()
{
    return m_parent;
}

/****************************************************************************/
/*! Sets a new \e parent for the QAxelObject. This method should only
    be used by subclasses when assigning a parent to a non-parented
    QAxelObject.

    \arg QAxelObject * Parent - the new parent.
*/
/****************************************************************************/
void QAxelObject::setParent (QAxelObject * parent) 
{
    m_parent = parent;
}

/****************************************************************************/
/*! Adds a new child to this group. The new child is appended last
    among the children.

    \arg QAxelObject * child - new child.
*/
/****************************************************************************/
void QAxelObject::addChild(QAxelObject * child) 
{  
    m_children.insert(m_children.end(), child) ;
    child->setParent(this) ;
}

/****************************************************************************/
/*! Adds a new child to this group. The new child is inserted \e
    before the child at \a Index. If \a Index is larger then the
    number of children the new child is put last.

    \arg QAxelObject * child - new child.  
    \arg const int unsigned index - the desired index.
*/
/****************************************************************************/
void QAxelObject::addChild (QAxelObject * child, const unsigned int index) 
{
    if(index > (uint)m_children.size())
	m_children.insert(m_children.end(), child) ;
    else
 	m_children.insert(m_children.begin()+index, child) ;

    child->setParent(this) ;
}

/****************************************************************************/
/*! Removes a child from the internal list of children. The child is
    returned if found, otherwise \c 0 is returned. If the child is
    found its reference-count is also decreased. It's up to the caller
    to delete the item (if desired) if the reference-count is \c 0.

    \arg QAxelObject * child - the child to be removed.
*/
/****************************************************************************/
QAxelObject * QAxelObject::removeChild (QAxelObject * child) 
{  
    int i = m_children.indexOf(child) ;
    if(i != -1)
	return m_children.takeAt(i) ;
    else 
	return 0 ;
}

/****************************************************************************/
/*! Removes all children from the group. They are deleted if \a Delete
    is \c true and the reference-count is \c 0, otherwise they are
    only dereferenced and not deleted even if the reference-count is
    \c 0.

	\warning If you set \a delete to \c false you had better know
	what you're doing, as it may lead to memeory-leaks if the
	dangling unparented QAxelObjects are not deleted!

    \arg const bool Delete = true - wether to delete unreferenced
    QAxelObjects.
*/
/****************************************************************************/
void QAxelObject::removeAllChildren (const bool deleteAll) 
{
    for(int index = 0 ; index < m_children.size() ; index++)
	if(deleteAll)
	    delete m_children[index];
    
    m_children.clear ();
}

/****************************************************************************/
/*! Attempts to find the child \a child in the list of
    children. Returns the child if found, otherwise \c 0. Checks its
    children recursively.

    \arg QAxelObject * child - the child to be found.
*/
/****************************************************************************/
QAxelObject * QAxelObject::findChild (QAxelObject * child) 
{
    if(this == child)
	return this ;

    foreach(QAxelObject * o, m_children)
	if(o->findChild(child) != 0)
	  return o ;

    return 0 ;
}
    
} ; // namespace QAxel
