/****************************************************************************
 * A x e l
 *****************************************************************************
 * QRationalSurface
 * 2007-05-22
 * Julien Wintz
 * Ashish Reddy
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :
 *
 ****************************************************************************/

# include <QMessageBox>

# include <QAxel/QBoundingBox.h>
# include <QAxel/QPoint3f.h>
# include <QAxel/QCurve.h>
# include <QAxel/QMesh.h>
# include <QAxel/PluginManager.h>

# include <QSisl/QSislInterface.h>
# include <QSynaps/QSynapsInterface.h>
# include <QCore/QCoreInterface.h>

namespace QAxel {

int QBSplineCurve::id_ = 0 ;

QBSplineCurve::QBSplineCurve(void) : QParametricCurve() 
{
    ikind = 1 ;
    idim = 3 ;
    icopy = 0 ;

    this->frame_ = new ManipulatedFrame() ;
    WorldConstraint * constraint = new WorldConstraint() ;
    constraint->setRotationConstraintType(AxisPlaneConstraint::FORBIDDEN) ;
    constraint->setTranslationConstraintType(AxisPlaneConstraint::FORBIDDEN) ;
    frame_->setConstraint(constraint) ;

    QString str ;
    setProperty("type", "BSplineCurve") ;
    setProperty("name", "bsc_"+str.setNum(id_)) ;
    id_++ ;
}


QBSplineCurve::QBSplineCurve(QBoundingBox * boundingbox) : QParametricCurve(boundingbox) 
{
    ikind = 1 ;
    idim = 3 ;
    icopy = 0 ;

    this->frame_ = new ManipulatedFrame() ;
    WorldConstraint * constraint = new WorldConstraint() ;
    constraint->setRotationConstraintType(AxisPlaneConstraint::FORBIDDEN) ;
    constraint->setTranslationConstraintType(AxisPlaneConstraint::FORBIDDEN) ;
    frame_->setConstraint(constraint) ;

    QString str ;
    setProperty("type", "BSplineCurve") ;
    setProperty("name", "bsc_"+str.setNum(id_)) ;
    id_++ ;
}

QPoint3f * QBSplineCurve::eval(float u)
{
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    if(iCore) return iCore->eval(this, u) ;
    else      return NULL ;
}

int QBSplineCurve::getDimension(void)
{
    return idim;
}

int QBSplineCurve::getNumber(void)
{
    return number;
}

int QBSplineCurve::getOrder(void)
{
    return order;
}

double * QBSplineCurve::getKnots(void)
{
    return knots;
}

void QBSplineCurve::insert(int pos, QPoint3f * p)
{
    connect(p->frame(), SIGNAL(modified()), this, SLOT(compute())) ; 
    points.insert(pos, p) ;
}

void QBSplineCurve::push_point(QPoint3f * p)
{
    connect(p->frame(), SIGNAL(modified()), this, SLOT(compute())) ; 
    points.push_back(p) ;
}

void QBSplineCurve::push_front(QPoint3f * p)
{
    connect(p->frame(), SIGNAL(modified()), this, SLOT(compute())) ; 
    points.push_front(p) ;
}

void QBSplineCurve::computeKnots(void)
{
    number = points.size() ;
    knots = new double[number + order] ;
    int knot_number = 1 ;
    
    for(int i = 0 ; i < order ; i++ )
	knots[i] = 0 ;
    
    for(int i = order ; i < number ; i++)
	knots[i] = knot_number++ ;
    
    for(int i = number ; i < number + order ; i++)
	knots[i] = knot_number ;

}

using namespace std ;

void QBSplineCurve::compute(void) 
{
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
    if(iSisl) iSisl->create(this) ;
    if(iCore) iCore->create(this) ;
    foreach(QPoint3f * p, points) { p->setSize(0.015) ; p->setRGB(119,136,153) ; }
}

void QBSplineCurve::draw(void) 
{
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    
    if(hidden()) return ;
    glPushMatrix() ;
    glMultMatrixd(frame_->matrix()) ;
    glColor() ;
    if(iCore) iCore->draw(this) ;
    glPopMatrix() ;

    if(mode() == QAxelObject::Edit || mode() == QAxelObject::Extrude) {
	foreach(QPoint3f * p, points)
	    p->draw() ;
	
	// Temporary control polygin drawing function
	glDisable(GL_LIGHTING) ;
	glColor3ub(200, 200, 200) ;
	for(int i = 0 ; i < points.size()-1 ; i++) {
	    glBegin(GL_LINES) ;
	    glVertex3f(points[i]->x(), points[i]->y(), points[i]->z()) ;
	    glVertex3f(points[i+1]->x(), points[i+1]->y(), points[i+1]->z()) ;
	    glEnd() ;
	}
	glEnable(GL_LIGHTING) ;
    }
    
    if(bbdrawn_) boundingbox_->draw() ;
}

void QBSplineCurve::drawWithNames(void) 
{
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    
    glPushName((GLuint)this);
    
    if(hidden()) return ;
    glPushMatrix() ;
    glMultMatrixd(frame_->matrix()) ;
    glColor() ;
    if(iCore) iCore->draw(this) ;
    glPopMatrix() ;
    glPopName();
    
    if(mode() == QAxelObject::Edit || mode() == QAxelObject::Extrude)
	foreach(QPoint3f * p, points)
	    p->drawWithNames() ;
}

} ;
