/****************************************************************************
 * A x e l
 *****************************************************************************
 * QBSplineSurface.cpp
 * 2007-05-22
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :
 *
 ****************************************************************************/

# include <QMessageBox>

# include <QAxel/QPoint3f.h>
# include <QAxel/QBoundingBox.h>
# include <QAxel/QSingularity.h>
# include <QAxel/QSurface.h>
# include <QAxel/QMesh.h>
# include <QAxel/PluginManager.h>

# include <QSisl/QSislInterface.h>
# include <QSynaps/QSynapsInterface.h>
# include <QCore/QCoreInterface.h>

using namespace std ;

namespace QAxel {

/*****************************************************************************
 * QBSplineSurface
 *****************************************************************************/

int QBSplineSurface::id_ = 0 ;

QBSplineSurface::QBSplineSurface(QBoundingBox * boundingbox) : QParametricSurface(boundingbox) 
{
    ikind = 1 ;
    idim = 3 ;
    icopy = 0 ;
    bbdrawn_ = false ;
    QString str ;
    setProperty("type", "BSplineSurface") ;
    setProperty("name", "bs_"+str.setNum(id_++)) ;
    setObjectColor(new QColor(255, 255, 54)) ;
    WorldConstraint * constraint = new WorldConstraint() ;
    constraint->setRotationConstraintType(AxisPlaneConstraint::FORBIDDEN) ;
    constraint->setTranslationConstraintType(AxisPlaneConstraint::FORBIDDEN) ;
    frame_->setConstraint(constraint) ;
}

void QBSplineSurface::compute(void) 
{	
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
    if(iCore) iCore->create(this) ;
    if(iSisl) iSisl->create(this) ;
    foreach(QPoint3f * p, points) { p->setSize(0.015) ; p->setRGB(119,136,153) ; }
}

void QBSplineSurface::draw(void)
{
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    if(hidden()) return ;
    glPushMatrix() ;
    glMultMatrixd(frame_->matrix()) ;
    glColor() ;
    if(iCore) iCore->draw(this) ;
    glPopMatrix() ;

    if(mode() == QAxelObject::Edit || mode() == QAxelObject::Extrude) {
	foreach(QPoint3f * p, points)
	    p->draw() ;
	
	// Temporary control polygin drawing function
	glDisable(GL_LIGHTING) ;
	glColor3ub(200, 200, 200) ;
	for(int i = 0 ; i < points.size()-1 ; i++) {
	    if( (i + 1) % number1 != 0) {
		glBegin(GL_LINES) ;
		glVertex3f(points[i]->x(), points[i]->y(), points[i]->z()) ;
		glVertex3f(points[i+1]->x(), points[i+1]->y(), points[i+1]->z()) ;
		glEnd() ;
	    }
	}
	for(int i = 0 ; i < points.size()-1 ; i++) {
	    if( i < number1*(number2-1)) {
		glBegin(GL_LINES) ;
		glVertex3f(points[i]->x(), points[i]->y(), points[i]->z()) ;
		glVertex3f(points[i+number1]->x(), points[i+number1]->y(), points[i+number1]->z()) ;
		glEnd() ;
	    }
	}
	glEnable(GL_LIGHTING) ;
    }
    
    if(bbdrawn_) boundingbox_->draw() ;
}

void QBSplineSurface::drawWithNames(void) 
{
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    
    glPushName((GLuint)this);
    
    if(hidden()) return ;
    glPushMatrix() ;
    glMultMatrixd(frame_->matrix()) ;
    glColor() ;
    if(iCore) iCore->draw(this) ;
    glPopMatrix() ;
    glPopName();
    
    if(mode() == QAxelObject::Edit || mode() == QAxelObject::Extrude)
	foreach(QPoint3f * p, points)
	    p->drawWithNames() ;

}
    
void QBSplineSurface::push_point(QPoint3f * p) 
{
    connect(p->frame(), SIGNAL(modified()), this, SLOT(compute())) ;
    points.push_back(p) ;
}

void QBSplineSurface::push_point(QPoint3f * p, int position) 
{
    connect(p->frame(), SIGNAL(modified()), this, SLOT(compute())) ;
    points.insert(position, p) ;
}

void QBSplineSurface::computeKnots1(void)
{

    knots1 = new double[number1 + order1] ;
    int knot_number = 1 ;
    
    for(int i = 0 ; i < order1 ; i++ )
	knots1[i] = 0 ;
    
    for(int i = order1 ; i < number1 ; i++)
	knots1[i] = knot_number++ ;
    
    for(int i = number1 ; i < number1 + order1 ; i++)
	knots1[i] = knot_number ;

}

void QBSplineSurface::computeKnots2(void)
{

    knots2 = new double[number2 + order2] ;
    int knot_number = 1 ;
    
    for(int i = 0 ; i < order2 ; i++ )
	knots2[i] = 0 ;
    
    for(int i = order2 ; i < number2 ; i++)
	knots2[i] = knot_number++ ;
    
    for(int i = number2 ; i < number2 + order2 ; i++)
	knots2[i] = knot_number ;

}

QList<QPoint3f *> QBSplineSurface::listSelectedControlPoints(void)
{
    QList<QPoint3f *> l ;
    
    foreach(QPoint3f * p, points)
	if(p->selected())
	    l << p ;

    return l ;
}

QBSplineSurface::ExtrusionEdge QBSplineSurface::extrusionPossible(void)
{
    QList<QPoint3f *> l = listSelectedControlPoints() ;

    bool up = true, down = true, left = true, right = true ;
    
    for(int i = 0 ; i < number1 ; i++)
	if(l[i] != points[i])
	    up = false ;
    if(up)
	return UP ;

    for(int i = 0 ; i < number1 ; i++)
	if(l[i] != points[number1*(number2-1) + i])
	    down = false ;
    if(down)
	return DOWN ;

    for(int i = 0 ; i < number2 ; i++)
	if(l[i] != points[i*number1])
	    left = false ;
    if(left)
	return LEFT ;

    for(int i = 0 ; i < number2 ; i++)
	if(l[i] != points[(i+1)*number1 - 1])
	    right = false ;
    if(right)
	return RIGHT ;

    return NONE ;
}

void QBSplineSurface::extrudeUpSide(void)
{
    for(int i = number1 - 1; i >= 0; i--) {
	QPoint3f * np = new QPoint3f(points[i+i]->x(), points[i+i]->y(), points[i+i]->z()) ;
//	connect(this, SIGNAL(mouseMoved(QPoint3f *)), np, SLOT(translate(QPoint3f *))) ;
	points.push_front(np) ;
	//addobject to selection
    }

    number2++ ;
}

void QBSplineSurface::extrudeDownSide(void)
{
    for(int i = 0; i < number1; i++) {
	QPoint3f * np = new QPoint3f(points[i+i]->x(), points[i+i]->y(), points[i+i]->z()) ;
//	connect(this, SIGNAL(mouseMoved(QPoint3f *)), np, SLOT(translate(QPoint3f *))) ;
	points.push_back(np) ;
	//addobject to selection
    }
    
    number2++ ;
}

void QBSplineSurface::extrudeLeftSide(void)
{
    for(int i = 0; i < number2; i++) {
	QPoint3f * np = new QPoint3f(points[number1*i+i]->x(), points[number1*i+i]->y(), points[number1*i+i]->z()) ;
//	connect(this, SIGNAL(mouseMoved(QPoint3f *)), np, SLOT(translate(QPoint3f *))) ;
	points.insert(number1*i+i, np) ;
	//addobject to selection
    }

    number1++ ;
}

void QBSplineSurface::extrudeRightSide(void)
{
    for(int i = 0; i < number2; i++) {
	QPoint3f * np = new QPoint3f(points[i+i]->x(), points[i+i]->y(), points[i+i]->z()) ;
//	connect(this, SIGNAL(mouseMoved(QPoint3f *)), np, SLOT(translate(QPoint3f *))) ;
	points.insert(number1*(i+1)+i, np) ;
	//addobject to selection
    }

    number1++ ;
}

void QBSplineSurface::getGlData(int * nbi, int * indices, int * nbv, double * vertices)
{
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    if(iCore) iCore->getGlData(this, nbi, indices, nbv, vertices) ;
}

QPoint3f * QBSplineSurface::eval(float u, float v) 
{
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    if(iCore) return iCore->eval(this, u, v) ;
    else      return NULL ;
}

} ;
