/*****************************************************************************
 *
 ****************************************************************************/

/*****************************************************************************
 * Includes
 ****************************************************************************/
#include <assert.h>
#include <QAxel/QBoundingSphere.h>

/*****************************************************************************
 * Are we using namespaces?
 ****************************************************************************/
namespace QAxel {

/****************************************************************************/
/*! Creates an empty bounding sphere centered around (0,0,0) and with the
    radius 0.
*/
/****************************************************************************/
QBoundingSphere::QBoundingSphere () {

  // set the radius
  m_Radius = 0;
}


/****************************************************************************/
/*! Creates a new bounding sphere with the given \a Radius. If \a Radius is
    negative the boundign sphere is invalid.

	\arg const Vec & Center - center of the bounding sphere
    \arg const float Radius - new radius.
*/
/****************************************************************************/
QBoundingSphere::QBoundingSphere (const Vec & Center, const float Radius) {

  // precautions
  assert ( Radius >= 0 );

  // store new radius and center
  m_Center = Center;
  m_Radius = Radius;
}

/****************************************************************************/
/*! Adds the new bounding sphere \a V to this. It is enlarged if necessary.

    \arg const QBoundingSphere & V - added sphere.
*/
/****************************************************************************/
void QBoundingSphere::add (const QBoundingSphere & BS) {

  // get distance from the center to the center of the added sphere
  float Distance = (m_Center-BS.center() ).norm();

  // is the added sphere totally inside this?
  if ( Distance + BS.radius () > m_Radius ) {
	// no, so we need to enlarge this sphere a bit
	m_Radius = Distance + BS.radius ();
  }
}


/****************************************************************************/
/*! Adds the new \link Vec vertex \endlink to this bounding sphere. The
    sphere is enlarged if the new vertex is outside it, i.e. the radius is
	increased so that the sphere engulfs \a V.

    \arg const Vec & V - new vertex to add.
*/
/****************************************************************************/
void QBoundingSphere::add (const Vec & V) {

  // get the distance from the center to the new vertex
  float Distance = (m_Center-V).norm();

  // is it larger then the old radius?
  if ( Distance > m_Radius ) {
	// yep, so it is the new radius
	m_Radius = Distance;
  }
}


/****************************************************************************/
/*! Adds the vertex created by \a X, \a Y and \a Z to this bounding sphere.
    The sphere is enlarged if the new vertex is outside it, i.e. the radius is
	increased so that the sphere engulfs the vertex.
	
    \arg const float X - x-coordinate of added vertex.
    \arg const float Y - y-coordinate of added vertex.
    \arg const float Z - z-coordinate of added vertex.
*/
/****************************************************************************/
void QBoundingSphere::add (const float X, const float Y, const float Z) {

  // get the distance from the center to the new vertex
  float Distance = (m_Center - Vec( X, Y, Z )).norm();

  // is it larger then the old radius?
  if ( Distance > m_Radius ) {
	// yep, so it is the new radius
	m_Radius = Distance;
  }
}


/****************************************************************************/
/*! Returns the \e radius of this bounding sphere.
*/
/****************************************************************************/
float QBoundingSphere::radius () const {

  // return desired data
  return m_Radius;
}


} ; // namespace QAxel

