/*****************************************************************************
 * A x e l
 *****************************************************************************
 * QMesh
 * 2006-03-25
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :         
 *
 ****************************************************************************/

# include <QAxel/QDisplayList.h>
# include <QAxel/QVertex.h>
# include <QAxel/QEdge.h>
# include <QAxel/QFace.h>
# include <QAxel/QMesh.h>

namespace QAxel {

int QMesh::id_ = 0 ;

QMesh::QMesh(void) : QAxelObject(0)
{                 	
    QString str ;
    setProperty("type", "Mesh") ;
    setProperty("name", "Mesh"+str.setNum(id_)) ;
    
    id_++ ;     
    setMode(None) ;
    setEditionMode(Vertex) ;
    setObjectColor(new QColor(185,185,185)) ;

    drawEdges = false ;
    edgesList = new QDisplayList ;
}

QMesh::QMesh(QMesh * m) : QAxelObject(0)
{
    QString str ;
    setProperty("type", "Mesh") ;
    setProperty("name", "Mesh"+str.setNum(id_)) ; 
    
    id_++ ;     
    setMode(None) ;	
    setEditionMode(Vertex) ;
    setObjectColor(new QColor(185,185,185)) ;	
    
    this->nindexes_ = m->nindexes_ ;
    this->nvertices_ = m->nvertices_ ;
    
    glindexes_ = new GLuint[nindexes_] ;  
    glvertices_ = new float[3*nvertices_] ;
    glnormals_ = new float[3*nvertices_] ;
    
    for(int i=0 ; i<m->vertices_.size() ; i++) {
	QVertex * v = new QVertex(
	    m->vertices_[i]->index(), 
	    m->vertices_[i]->x(),
	    m->vertices_[i]->y(),
	    m->vertices_[i]->z(),
	    this) ;
	this->push_vertex(v) ;
    }
    
    for(int i=0 ; i<m->edges_.size() ; i++) {
	QEdge * e = new QEdge(m->edges_[i]->index1, m->edges_[i]->index2, this) ;
	push_edge(e) ;
    }
    
    for(int i=0 ; i<m->faces_.size() ; i++) {
	QFace * f = new QFace(m->faces_[i]->index(), this) ;
	for(int j=0 ; j<m->faces_[i]->indexes_.size() ; j++) f->push_index(m->faces_[i]->indexes_[j]) ;
	push_face(f) ;
	
	Vec norm = f->computeNormal() ;
	for(int i = 0 ; i < f->indexes_.size() ; i++) vertices_[f->indexes_[i]]->addToNormal(norm.x, norm.y, norm.z) ;
    }
    
    if(faces_.size()>0) for(int i=0 ; i<nvertices_ ; i++) {
	Vec v(this->glnormals_[3*i+0], this->glnormals_[3*i+1], this->glnormals_[3*i+2]) ;   
	v.normalize() ;
	this->glnormals_[3*i+0] = v.x ;
	this->glnormals_[3*i+1] = v.y ;
	this->glnormals_[3*i+2] = v.z ;
    }

    drawEdges = false ;
    edgesList = new QDisplayList ;
}

QMesh::QMesh(int nvertices, int nfaces) : QAxelObject(0)
{
    this->nindexes_ = nfaces*3 ;
    this->nvertices_ = nvertices ;
    
    glindexes_ = new GLuint[nindexes_] ;  
    glvertices_ = new float[3*nvertices_] ;
    glnormals_ = new float[3*nvertices_] ;
    for(int i = 0 ; i < nvertices_*3 ; i++) glnormals_[i] = 0.0f ;
    
    QString str ;
    setProperty("type", "Mesh") ;
    setProperty("name", "mesh_"+str.setNum(id_)) ; 
    
    id_++ ;     
    setMode(None) ;	
    setEditionMode(Vertex) ;
    setObjectColor(new QColor(185,185,185)) ;

    drawEdges = false ;
    edgesList = new QDisplayList ;
}

void QMesh::setNumberOfVertices(int n) 
{
    nvertices_ = n ;
    glvertices_ = new float[n*3] ;
    glnormals_ = new float[n*3] ;
    for(int i = 0 ; i < n*3 ; i++) glnormals_[i] = 0.0f ;
}

void QMesh::setNumberOfEdges(int n)
{
    nindexes_ = n*2 ;
    glindexes_ = new GLuint[n*2] ;    
}

void QMesh::setNumberOfFaces(int n) 
{
    nindexes_ = n*3 ;    
    glindexes_ = new GLuint[n*3] ;    
}

QList<QVertex *>& QMesh::vertices(){
    return vertices_ ;  
}

QList<QEdge *>& QMesh::edges(){
    return edges_ ;
}

QList<QFace *>& QMesh::faces(){
    return faces_ ;
}


double * QMesh::getVertex(int index) 
{ 
    double * v = new double[3] ;
    v[0] = glvertices_[3*index+0] ;
    v[1] = glvertices_[3*index+1] ;
    v[2] = glvertices_[3*index+2] ; 
    return v ;
}

void QMesh::setVertexNormal(int index, double nx, double ny, double nz) 
{
    glnormals_[3*index+0] = nx ;
    glnormals_[3*index+1] = ny ;
    glnormals_[3*index+2] = nz ;
}

void QMesh::computeNormals(void)
{
    for(int i = 0 ; i < faces_.size() ; i++) {
	Vec norm = faces_[i]->computeNormal() ;
	for(int j = 0 ; j < faces_[i]->indexes_.size() ; j++)
	    vertices_[faces_[i]->indexes_[j]]->addToNormal(norm.x, norm.y, norm.z) ;	
    }
	
    for(int i=0 ; i<nvertices_ ; i++) {
	Vec v(glnormals_[3*i+0], glnormals_[3*i+1], glnormals_[3*i+2]) ;   
	v.normalize() ;
	glnormals_[3*i+0] = v.x ;
	glnormals_[3*i+1] = v.y ;
	glnormals_[3*i+2] = v.z ;
    }
}

void QMesh::draw(void) 
{	
    if(hidden()) return ;
    glPushMatrix() ;     
    glMultMatrixd(frame_->matrix()) ;
    if(mode() == Edit) {              
	// First we draw vertices
	glDisable(GL_LIGHTING) ;
	glPointSize(3.0) ;
	for(int i=0 ; i<vertices_.size() ; i++) vertices_[i]->draw() ;
	glPointSize(1.0) ;     
	for(int i = 0 ; i < edges_.size() ; i++) edges_[i]->draw() ;
	glEnable(GL_LIGHTING) ;
	if(meshType() == Surface) {
	    for(int i = 0 ; i < faces_.size() ; i++) faces_[i]->draw() ;
	}
    } else {                
	if(meshType() == Surface) {                                                  
	    glColor() ;                               
	    glEnableClientState(GL_VERTEX_ARRAY) ;
	    glEnableClientState(GL_NORMAL_ARRAY) ;
	    glVertexPointer(3, GL_FLOAT, 0, glvertices_) ;
	    glNormalPointer(GL_FLOAT, 0, glnormals_) ;
	    glDrawElements(GL_TRIANGLES, nindexes_, GL_UNSIGNED_INT, glindexes_) ;
	    glDisableClientState(GL_NORMAL_ARRAY) ;
	    glDisableClientState(GL_VERTEX_ARRAY) ;

	    if(drawEdges) {
	    if(edgesList->isValid())
		edgesList->draw() ;
	    else {
		edgesList->begin() ;
		glDisable(GL_LIGHTING) ;
		glColor3ub(100,100,100) ;
		for(int i = 0 ; i < edges_.size() ; i++) edges_[i]->draw() ;
		glEnable(GL_LIGHTING) ;
		edgesList->end() ;
		edgesList->draw() ;
	    }
	    }
	} else {
	    glDisable(GL_LIGHTING) ;
	    glColor() ;
	    for(int i = 0 ; i < edges_.size() ; i++) edges_[i]->draw() ;
	    glEnable(GL_LIGHTING) ;
	}
    }                                                                             
    glPopMatrix() ;                      
}                                                               

void QMesh::drawWithNames(void) 
{
    if(hidden()) return ;
    glPushMatrix() ;     
    glMultMatrixd(frame_->matrix()) ;
    if(mode() == Edit) {              
	glPointSize(3.0) ;
	if(editionMode_ == Vertex) for(int i=0 ; i<vertices_.size() ; i++) vertices_[i]->drawWithNames() ;
	glPointSize(1.0) ;
	if(editionMode_ == Edge) for(int i = 0 ; i < edges_.size() ; i++) edges_[i]->drawWithNames() ;
	if(editionMode_ == Face) for(int i = 0 ; i < faces_.size() ; i++) faces_[i]->drawWithNames() ;
    } else {                
	glPushName((GLuint)this);
	if(meshType() == Surface) {                      
	    glColor() ;                               
	    glEnableClientState(GL_VERTEX_ARRAY) ;
	    glEnableClientState(GL_NORMAL_ARRAY) ;
	    glVertexPointer(3, GL_FLOAT, 0, glvertices_) ;
	    glNormalPointer(GL_FLOAT, 0, glnormals_) ;
	    glDrawElements(GL_TRIANGLES, nindexes_, GL_UNSIGNED_INT, glindexes_) ;
	    glDisableClientState(GL_NORMAL_ARRAY) ;
	    glDisableClientState(GL_VERTEX_ARRAY) ;
	} else {
	    glDisable(GL_LIGHTING) ;
	    glColor() ;
	    for(int i = 0 ; i < edges_.size() ; i++) edges_[i]->draw() ;
	    glEnable(GL_LIGHTING) ;
	}
	glPopName() ;
    }                                                                             
    glPopMatrix() ;                      
}

void QMesh::drawWithoutColor(void) 
{
    if(hidden()) return ;
    glPushMatrix() ;     
    glMultMatrixd(frame_->matrix()) ;
    if(mode() == Edit) {              
	// First we draw vertices
	glDisable(GL_LIGHTING) ;
	glPointSize(3.0) ;
	for(int i=0 ; i<vertices_.size() ; i++) vertices_[i]->draw() ;
	glPointSize(1.0) ;     
	for(int i = 0 ; i < edges_.size() ; i++) edges_[i]->draw() ;
	glEnable(GL_LIGHTING) ;
	if(meshType() == Surface) {
	    for(int i = 0 ; i < faces_.size() ; i++) faces_[i]->draw() ;
	}
    } else {                
	if(meshType() == Surface) {                                                  
// 			glColor() ;                               
	    glEnableClientState(GL_VERTEX_ARRAY) ;
	    glEnableClientState(GL_NORMAL_ARRAY) ;
	    glVertexPointer(3, GL_FLOAT, 0, glvertices_) ;
	    glNormalPointer(GL_FLOAT, 0, glnormals_) ;
	    glDrawElements(GL_TRIANGLES, nindexes_, GL_UNSIGNED_INT, glindexes_) ;
	    glDisableClientState(GL_NORMAL_ARRAY) ;
	    glDisableClientState(GL_VERTEX_ARRAY) ;
	} else {
	    glDisable(GL_LIGHTING) ;
// 			glColor() ;
	    for(int i = 0 ; i < edges_.size() ; i++) edges_[i]->draw() ;
	    glEnable(GL_LIGHTING) ;
	}
    }                                                                             
    glPopMatrix() ;                      
}     

void QMesh::clear(void) 
{
    vertices_.clear() ;
    edges_.clear() ;
    faces_.clear() ;
}

void QMesh::translate(float x, float y, float z)
{
    // for(int i=0 ; i < nvertices_ ; i++) {
    // 	glvertices_[3*i+0]+=x ;
    // 	glvertices_[3*i+1]+=y ;
    // 	glvertices_[3*i+2]+=z ;
    // }
    frame()->setPosition(frame()->position().x + x, frame()->position().y + y, frame()->position().z + z) ;
}

} ; // namespace QAxel
