/*****************************************************************************
 * A x e l                                                                   *
 ******************************************************************************
 * QRegionEdge                                                               *
 * 2006-10-16                                                                *
 * Julien Wintz                                                              *
 ******************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis                   *
 ******************************************************************************
 * Comments :                                                                *
 *****************************************************************************/
 
# include <QData/QRegionEdge.h>

namespace QData {

QRegionEdge::QRegionEdge(QRegionVertex * p1, QRegionVertex * p2, QAxelObject * o) 
{
    e_source = p1 ;
    e_dest = p2 ;
    e_object = o ;
}

QRegionEdge::~QRegionEdge(void) 
{

}

/*!************************************************************************************
 * Returns true if \param r is an extremity of the edge.
 **************************************************************************************/
bool QRegionEdge::contains(QRegionVertex * r)
{
    return (e_source == r) || (e_dest == r) ;
}

QRegionVertex * QRegionEdge::other(QRegionVertex * v)
{
    return (v == e_source) ? e_dest : e_source ;
}

QAxelObject * QRegionEdge::object(void) 
{
    return e_object ; 
}

QRegionVertex * QRegionEdge::source(void)
{ 
    return e_source ; 
}

QRegionVertex * QRegionEdge::dest(void)
{ 
    return e_dest ; 
}

QString QRegionEdge::type(void)
{
    return e_object->type() ;
}

QList<QRegionVertex *> QRegionEdge::commonVerticesOf(QRegionEdge * e)
{
    QList<QRegionVertex *> l ;
    if(e_source->equals(e->source())) l << e->source() ;
    if(e_source->equals(e->dest())) l << e->dest() ;
    if(e_dest->equals(e->source())) l << e->source() ;
    if(e_dest->equals(e->dest())) l << e->dest() ;
    return l ;
}

QList<QRegionVertex *> QRegionEdge::commonVerticesWith(QRegionEdge * e)
{
    QList<QRegionVertex *> l ;
    if(e_source->equals(e->source())) l << e_source ;
    if(e_source->equals(e->dest())) l << e_source ;
    if(e_dest->equals(e->source())) l << e_dest ;
    if(e_dest->equals(e->dest())) l << e_dest ;
    return l ;
}

bool QRegionEdge::equals(QRegionEdge *e) 
{
    QRegionEdge * e1 = this ;
    QRegionEdge * e2 = e ;
    if(e1->source()->equals(e2->source()) && e1->dest()->equals(e2->dest()) && e1->object()->type() == e2->object()->type()) return true ;
    if(e1->source()->equals(e2->dest()) && e1->dest()->equals(e2->source()) && e1->object()->type() == e2->object()->type()) return true ;
    if(e1->dest()->equals(e2->source()) && e1->source()->equals(e2->dest()) && e1->object()->type() == e2->object()->type()) return true ;
    if(e1->dest()->equals(e2->dest()) && e1->source()->equals(e2->source()) && e1->object()->type() == e2->object()->type()) return true ;
    return false ;
}

bool QRegionEdge::alignedWith(QRegionEdge * e)
{
    if(e_source->x() == e_dest->x() && e_dest->x() == e->source()->x() && e->source()->x() == e->dest()->x()) return true ;
    if(e_source->y() == e_dest->y() && e_dest->y() == e->source()->y() && e->source()->y() == e->dest()->y()) return true ;
    return false ;
}

bool QRegionEdge::isCellBorder(void) 
{
    return e_object->type() == "Cell" ;
}

bool QRegionEdge::mergesWith(QRegionEdge * e)
{
    QList<QRegionVertex *> commons = commonVerticesOf(e) ;
    
    if(commons.size() == 0) 
	return false ;
    
    if(alignedWith(e) && this->isCellBorder() && e->isCellBorder())
	return true ;
    
    return false ;
}

/*!************************************************************************************
 * Merges the current edge with \param e and returns the list of resulting unhooked 
 * vertices.
 **************************************************************************************/
QList<QRegionVertex *> QRegionEdge::mergeWith(QRegionEdge * e)
{
    QList<QRegionVertex *> unhooked ;
    QRegionVertex * common = commonVerticesOf(e).first() ;
    
//     if(common->equals(e_source))
// 	if(common->equals(e->source())) { e_source = e_dest ; e_dest = e->dest() ; unhooked << e->source() ; } 
// 	else                            { e_source = e_dest ; e_dest = e->source() ; unhooked << e->dest() ; }
//     else
// 	if(common->equals(e->source())) { e_dest = e->dest() ; unhooked << e->source() ; }
// 	else 		                { e_dest = e->source() ; unhooked << e->dest() ; }

    if(common->equals(e_source))
	if(common->equals(e->source())) { e_source = e->dest() ; unhooked << e->source() ; } 
	else                            { e_source = e->source() ; unhooked << e->dest() ; }
    else
	if(common->equals(e->source())) { e_dest = e->dest() ; unhooked << e->source() ; }
	else 		                { e_dest = e->source() ; unhooked << e->dest() ; }
    
    return unhooked ;
}

void QRegionEdge::replace(QRegionVertex * p1, QRegionVertex * p2)
{
    if(e_source == p1) e_source = p2 ;
    if(e_dest   == p1) e_dest   = p2 ;
}

void QRegionEdge::switchVertices(void)
{
    QRegionVertex * tmp = e_source ;
    e_source = e_dest ;
    e_dest = tmp ;
}

QString QRegionEdge::toString(void)
{
    QString text = "(" + e_source->toString() + ", " + e_dest->toString() + ", " + e_object->name() + ", " + e_object->type() + ")" ;
    return text ;
}

std::ostream& operator << (std::ostream& stream, QRegionEdge & e) 
{ 
    return stream << e.toString().toLatin1().data() ; 
}

std::ostream& operator << (std::ostream& stream, QRegionEdge * e) 
{ 
    return stream << e->toString().toLatin1().data() ; 
}

QBoundingBox * QRegionEdge::box(void)
{
//    if(e_box) return e_box ;

    float xmin = (e_source->x() <= e_dest->x()) ? e_source->x() : e_dest->x() ;
    float xmax = (e_source->x() >  e_dest->x()) ? e_source->x() : e_dest->x() ;
    float ymin = (e_source->y() <= e_dest->y()) ? e_source->y() : e_dest->y() ;
    float ymax = (e_source->y() >  e_dest->y()) ? e_source->y() : e_dest->y() ;

    e_box = new QBoundingBox(xmin, xmax, ymin, ymax) ;

    if(e_box->width() < 1e-2) {
	e_box->setXMin(e_box->xmin()-0.01) ;
	e_box->setXMax(e_box->xmax()+0.01) ;
    }

    if(e_box->height() < 1e-2) {
	e_box->setYMin(e_box->ymin()-0.01) ;
	e_box->setYMax(e_box->ymax()+0.01) ;
    }

    return e_box ;
}

QRegionEdge::MonotonyCode QRegionEdge::code(void)
{
    using namespace std ;
         if(e_source->x() <  e_dest->x() && e_source->y() <  e_dest->y()) return PlusPlus ;
    else if(e_source->x() <  e_dest->x() && e_source->y() >  e_dest->y()) return PlusMinus ;
    else if(e_source->x() >  e_dest->x() && e_source->y() <  e_dest->y()) return MinusPlus ;
    else if(e_source->x() >  e_dest->x() && e_source->y() >  e_dest->y()) return MinusMinus ;
    else if(e_source->x() == e_dest->x() && e_source->y() <  e_dest->y()) return NullPlus ;
    else if(e_source->x() == e_dest->x() && e_source->y() >  e_dest->y()) return NullMinus ;
    else if(e_source->x() <  e_dest->x() && e_source->y() == e_dest->y()) return PlusNull ;
    else if(e_source->x() >  e_dest->x() && e_source->y() == e_dest->y()) return MinusNull ;
    else if(e_source->x() == e_dest->x() && e_source->y() == e_dest->y()) return NullNull ;
    else return Undefined ;
}

QString QRegionEdge::codeString(void)
{
    using namespace std ;
         if(e_source->x() <  e_dest->x() && e_source->y() <  e_dest->y()) return "(+,+)" ;
    else if(e_source->x() <  e_dest->x() && e_source->y() >  e_dest->y()) return "(+,-)" ;
    else if(e_source->x() >  e_dest->x() && e_source->y() <  e_dest->y()) return "(-,+)" ;
    else if(e_source->x() >  e_dest->x() && e_source->y() >  e_dest->y()) return "(-,-)" ;
    else if(e_source->x() == e_dest->x() && e_source->y() <  e_dest->y()) return "(0,+)" ;
    else if(e_source->x() == e_dest->x() && e_source->y() >  e_dest->y()) return "(0,-)" ;
    else if(e_source->x() <  e_dest->x() && e_source->y() == e_dest->y()) return "(+,0)" ;
    else if(e_source->x() >  e_dest->x() && e_source->y() == e_dest->y()) return "(-,0)" ;
    else if(e_source->x() == e_dest->x() && e_source->y() == e_dest->y()) return "(0,0)" ;
    else return "(?,?)" ;
}

} ; // namespace QData
