/*****************************************************************************
 * A x e l
 *****************************************************************************
 * Viewer.h
 * 2006-03-31
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :
 ****************************************************************************/

# include <QFile>
# include <QCursor>
# include <QAction>
# include <QMenu>
# include <QPainter>

# include <QGui/Viewer.h>
# include <QGui/ManipulatedFrameSetConstraint.h>

# include <QP5Glove/QP5GloveInterface.h>

# include <QAxel/QVertex.h>
# include <QAxel/QEdge.h>
# include <QAxel/QFace.h>
# include <QAxel/QMesh.h>
# include <QAxel/PluginManager.h>
# include <QAxel/QBoundingBox.h>
# include <QAxel/QPlane.h>
# include <QAxel/QCurve.h>

# include <QGui/QDimensionModeWidget.h> 
# include <QGui/MainWindow.h>

# ifndef GL_MULTISAMPLE
# define GL_MULTISAMPLE  0x809D
# endif

using namespace std ;

Viewer::Viewer(QWidget *parent) : QGLViewer(parent, 0, 0) 
{
    selectionMode_ = NONE;      
    xyGridIsDrawn_ = true ;
    xzGridIsDrawn_ = true ;
    yzGridIsDrawn_ = true ;
    //************** AF SandBox
    directMode = false;
	
    if (directMode)
    {
	connect(this,SIGNAL(directObjectAdded(const Vec&)),parent,SLOT(addObject(const Vec&)));
    }
    //***********************AF
    setObjectName("Viewer");
    
    //setting camera constraints
    cameraConstraint = new CameraConstraint(camera()) ;
	
    // Setting up actions
    refineImplicitCurveAction = new QAction("Implicit Curve", this) ;
    connect(refineImplicitCurveAction, SIGNAL(triggered()), this, SLOT(refineImplicitCurves())) ;
	
    autoIntersectParametricSurfacesAction = new QAction("Auto-Intersect Parametric Surfaces", this) ;
    intersectRationalSurfacesAction = new QAction("Intersection Rational Surfaces (beta)", this) ;
    intersectParametricSurfacesAction = new QAction("Intersect Parametric Surfaces", this) ;
    intersectImplicitCurvesAction = new QAction("Intersect Implicit Curves", this) ;
    intersectImplicitSurfacesAction = new QAction("Intersect Implicit Surfaces", this) ;
    interpolatePointsAsBSplineAction = new QAction("Interpolate points as BSplineCurve", this) ;
    intersectBSplineCurvesAction = new QAction("Intersect BSpline Curves", this) ;
    offsetBSplineCurvesAction = new QAction("Offset BSpline Curves", this) ;
    blendBSplineCurvesAction = new QAction("Blend BSpline Curves", this) ;
    quadClipImplicitCurvesAction = new QAction("Clip Implicit Curves", this) ;
    approximatePointsAsBSplineAction = new QAction("Approximate points as BSplineCurve", this) ;
	
    connect(autoIntersectParametricSurfacesAction, SIGNAL(triggered()), this, SLOT(autoIntersectParametricSurfaces())) ;
    connect(intersectParametricSurfacesAction, SIGNAL(triggered()), this, SLOT(intersectParametricSurfaces())) ;
    connect(interpolatePointsAsBSplineAction, SIGNAL(triggered()), this, SLOT(interpolatePointsAsBSpline())) ;
    connect(intersectImplicitSurfacesAction, SIGNAL(triggered()), this, SLOT(intersectImplicitSurfaces())) ;
    connect(intersectBSplineCurvesAction, SIGNAL(triggered()), this, SLOT(intersectBSplineCurves())) ;
    connect(offsetBSplineCurvesAction, SIGNAL(triggered()), this, SLOT(offsetBSplineCurves())) ;
    connect(blendBSplineCurvesAction, SIGNAL(triggered()), this, SLOT(blendBSplineCurves())) ;
    connect(intersectRationalSurfacesAction, SIGNAL(triggered()), this, SLOT(intersectRationalSurfaces())) ;
    connect(intersectImplicitCurvesAction, SIGNAL(triggered()), this, SLOT(intersectImplicitCurves())) ;
    connect(quadClipImplicitCurvesAction, SIGNAL(triggered()), this, SLOT(quadClipImplicitCurves())) ;
    connect(approximatePointsAsBSplineAction, SIGNAL(triggered()), this, SLOT(approximatePointsAsBSplineCurve())) ;
    
    setAcceptDrops(true) ;
    setAttribute(Qt::WA_NoSystemBackground);

    // Connecting signals from the object manager.
    connect(ObjectManager::instance(), SIGNAL(objectAdded(QAxelObject *)),   this, SLOT(update())) ;
    connect(ObjectManager::instance(), SIGNAL(objectRemoved(QAxelObject *)), this, SLOT(update())) ;
}

void Viewer::setXYGridIsDrawn(bool b) { xyGridIsDrawn_ = b ; this->updateGL() ; }
void Viewer::setXZGridIsDrawn(bool b) { xzGridIsDrawn_ = b ; this->updateGL() ; }
void Viewer::setYZGridIsDrawn(bool b) { yzGridIsDrawn_ = b ; this->updateGL() ; }

void Viewer::init() 
{
    setShortcut(DRAW_AXIS, 0) ;
    setShortcut(DRAW_GRID, 0);
    setShortcut(EXIT_VIEWER, 0) ;
    setShortcut(SAVE_SCREENSHOT, 0) ;
    setShortcut(FULL_SCREEN, 0) ;
    setShortcut(HELP, 0) ;
    setShortcut(CAMERA_MODE, 0) ;
	
    //************** AF SandBox
    if(directMode)
    {
	setMouseBinding(Qt::LeftButton, CAMERA, ROTATE);
	setMouseBinding(Qt::RightButton, CAMERA, TRANSLATE);
	setWheelBinding(Qt::NoModifier, CAMERA, ZOOM);
    }
    //*************************

    // A ManipulatedFrameSetConstraint will apply displacements to the selection
    setManipulatedFrame(new ManipulatedFrame());
    manipulatedFrame()->setConstraint(new ManipulatedFrameSetConstraint());
  
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,GL_TRUE) ;   

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_MULTISAMPLE);
    glEnable(GL_LINE_SMOOTH) ;
    glEnable(GL_BLEND) ;

    glShadeModel(GL_SMOOTH) ;
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA) ; 
	
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL) ;
    glPolygonOffset(1.0, 1.0) ;
    glEnable(GL_POLYGON_OFFSET_POINT) ;
    glEnable(GL_POLYGON_OFFSET_LINE) ;
    glEnable(GL_POLYGON_OFFSET_FILL) ;

    setSceneRadius(20) ;

    QP5GloveInterface * iP5Glove = dynamic_cast<QP5GloveInterface *>(PluginManager::instance()->plugin("P5Glove")) ;
    if(iP5Glove) startAnimation() ;

    mf_move = false ;
    v_oldMousePosition = new QPoint3f(0.0, 0.0, 0.0) ;
}

void Viewer::setAntiAliasing(bool enable)
{
    // is antialisasing enabled? 
    if (enable) {
	// yep, so do the necessary setup
	glEnable    ( GL_LINE_SMOOTH);
	glEnable    ( GL_BLEND);
	glBlendFunc ( GL_SRC_ALPHA,        GL_ONE_MINUS_SRC_ALPHA );
	glHint      ( GL_LINE_SMOOTH_HINT, GL_DONT_CARE );
    }

    else {
	// disable antialiasing
	glDisable ( GL_LINE_SMOOTH);
	glDisable ( GL_BLEND);
    }
}

void Viewer::setSmoothShading(bool enable)
{
    if (enable) glShadeModel ( GL_SMOOTH ) ;
    else glShadeModel ( GL_FLAT ) ;
}

void Viewer::setDepthTest(bool enable)
{
    if (enable) glEnable ( GL_DEPTH_TEST );
    else glDisable ( GL_DEPTH_TEST );
}

void Viewer::setFaceCulling(bool enable)
{
    // is culling enabled? 
    if (enable) {
	// yep, so do the necessary setup
	// glCullFace ( (GLenum)m_Current );
	glEnable   ( GL_CULL_FACE );
    }

    else {
	// disable backface culling
	glDisable ( GL_CULL_FACE );
    }
}

QString Viewer::helpString() const 
{
    QString text = "<h2>Axel modeler controls</h2>" ;
    text += "This dialog summarizes the controls of Axel";
    return text;
}

// O b j e c t  M a n a g e m e n t  f u n c t i o n

void Viewer::addObject(QAxelObject * object) 
{
    if(!objects_.contains(object)) objects_ << object ;
    nameMap_.insert(object->name(), object) ;
    updateGL() ;
    emit objectAdded(object) ;
    emit sceneChanged();
}

void Viewer::addObjects(QList<QAxelObject *> objectList)
{
    for(int i=0 ; i < objectList.size() ; i++) {
	objects_ << objectList[i] ;
	emit objectAdded(objectList[i]) ;
    }
}

void Viewer::removeObject(QAxel::QAxelObject *object)
{
    objects_.removeAll(object);
    nameMap_.erase(nameMap_.find(object->name())) ;
    // delete object;
    emit objectRemoved() ;
    emit sceneChanged();
    updateGL() ;
}

void Viewer::removeSelectedObjects(void)
{
    for (int i = 0 ; i < selection_.size() ; i++)
	removeObject(selection_[i]) ;
    selection_.clear() ;
    updateGL() ;
}

QAxelObject * Viewer::getObjectFromName(QString name) 
{
    QAxelObject * o = NULL ;
    for(QList<QAxelObject *>::iterator it=objects_.begin(), end=objects_.end(); it != end; ++it)
    {
	// First, we take care about container objects
	if((*it)->type() == "Mesh")
	{
	    QMesh * m = (QMesh *)(*it) ;
	    if(m->mode() == QMesh::Edit) 
	    {
		for(int i=0 ; i<m->vertices_.size() ; i++)
		{
		    if(m->vertices_[i]->name() == name)
			o = (QAxelObject*)m->vertices_[i] ;
		}
		for(int i=0 ; i<m->edges_.size() ; i++)
		{
		    if(m->edges_[i]->name() == name)
			o = (QAxelObject*)m->edges_[i] ;
		}
		for(int i=0 ; i<m->faces_.size() ; i++)
		{
		    if(m->faces_[i]->name() == name)
			o = (QAxelObject*)m->faces_[i] ;
		}
	    }
	    if((*it)->name() == name) o = *it ;
	} else if((*it)->name() == name) o = *it ;
    }
    return o ;
}

QList<QAxelObject *> Viewer::listObjects(void) 
{
    return objects_ ;
}

QList<QAxelObject *> Viewer::listSelectedObjects(void)
{
    QList<QAxelObject *> l ;
    for(int i=0 ; i<selection_.size() ; i++)
    {
	l << selection_[i] ;
    }
    return l ;
}
QList<QAxelObject *> Viewer::listGrabbedObjects(void) 
{
    QList<QAxelObject *> l ;
    for(int i=0 ; i<selection_.size() ; i++) 
    {
	QAxelObject * o = selection_[i];
	if(o && o->mode() == QAxelObject::Grab)
	    l << o ;
    }
    return l ;
}

QList<QAxelObject *> Viewer::listExtrusionObjects(void) 
{
    QList<QAxelObject *> l ;
    for(int i=0 ; i<objects_.size() ; i++) 
    {
	QAxelObject * o = objects_[i];
	if(o && o->mode() == QAxelObject::Extrude)
	    l << o ;
    }
    return l ;
}

QList<QAxelObject *> Viewer::listObjectsFromType(QString type) 
{
    QList<QAxelObject *> l ;
    for(QList<QAxelObject *>::iterator it=objects_.begin(), end=objects_.end(); it != end; ++it)
	if((*it)->type() == type)
	    l << *it ;
    return l ;
}   

QList<QAxelObject *> Viewer::listSelectedObjectsFromType(QString type) 
{
    QList<QAxelObject *> l ;
    for(int i=0 ; i<selection_.size() ; i++) {
	QAxelObject * o = selection_[i];
	if(o && o->type() == type)
	    l << o ;
    }
    return l ;
}
QList<QAxelObject *> Viewer::listEditedObjectsFromType(QString type) 
{
    QList<QAxelObject *> l ;
    for(int i=0 ; i<objects_.size() ; i++) {
	QAxelObject * o = objects_[i];
	if(o && o->type() == type && o->mode() == QAxelObject::Edit)
	    l << o ;
    }
    return l ;
}
QList<QAxelObject *> Viewer::listExtrusionObjectsFromType(QString type) 
{
    QList<QAxelObject *> l ;
    for(int i=0 ; i<objects_.size() ; i++) 
    {
	QAxelObject * o = objects_[i];
	if(o && o->type() == type && o->mode() == QAxelObject::Extrude)
	    l << o ;
    }
    return l ;
}
QList<QAxelObject *> Viewer::listGrabbedObjectsFromType(QString type) 
{
    QList<QAxelObject *> l ;
    for(int i=0 ; i<selection_.size() ; i++) 
    {
	QAxelObject * o = selection_[i];
	if(o && o->type() == type && o->mode() == QAxelObject::Grab)
	    l << o ;
    }
    return l ;
}

//************** AF SandBox

void Viewer::selectFromName(QString objectName)
{
    addObjectToSelection(getObjectFromName(objectName));
    emit objectSelected(getObjectFromName(objectName)) ;
}

void Viewer::setSelectedObjects(QList<QAxelObject *> select)
{
    unselectAll();
    for(int i=0 ; i<select.size() ; i++)
    {
	addObjectToSelection(select.at(i));
    }
    updateGL();
}

QList<QString> Viewer::getListObjectsName()
{
    return nameMap_.keys();
}

void Viewer::setWireDisplay(void) 
{ 
    glDisable(GL_LIGHTING) ; 
    glShadeModel(GL_FLAT) ; 
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE) ; 
    updateGL() ;
}

void Viewer::setLitWireDisplay(void) 
{ 
    glEnable(GL_LIGHTING) ; 
    glShadeModel(GL_SMOOTH) ; 
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE) ; 
    updateGL() ;
}

void Viewer::setFlatDisplay(void) 
{ 
    glEnable(GL_LIGHTING) ; 
    glShadeModel(GL_FLAT) ; 
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL) ; 
    updateGL() ;
}

void Viewer::setSmoothDisplay(void) 
{ 
    glEnable(GL_LIGHTING) ; 
    glShadeModel(GL_SMOOTH) ; 
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL) ; 
    updateGL() ;
}

void Viewer::toggleMeshEditionMode(void)
{
    QList<QAxelObject *> l = listSelectedObjectsFromType("Mesh") ;
    if(l.size()==0) l = listObjectsFromType("Mesh") ;
    bool allToEdit = true ;
    for(int i=0 ; i < l.size() ; i++) {
	if(((QMesh *)l[i])->mode() == QMesh::Edit) allToEdit = false ;
    }                                           
    for(int i=0 ; i < l.size() ; i++) {
	if(allToEdit) 	{ 
	    ((QMesh *)l[i])->setMode(QMesh::Edit) ;        
	} else { 
	    ((QMesh *)l[i])->setMode(QMesh::None) ;     
	    QMesh * m = (QMesh *)l[i] ;
	    for(int i=0 ; i<m->vertices_.size() ; i++) m->vertices_[i]->deselect() ;
	    for(int i=0 ; i<m->edges_.size() ; i++) m->edges_[i]->deselect() ;
	    for(int i=0 ; i<m->faces_.size() ; i++) m->faces_[i]->deselect() ;
	    updateSelection() ;
	}
    }   
    updateGL() ;
}

void Viewer::selectAll(void) 
{
    unselectAll() ;
    for(int i=0 ; i<objects_.size() ; i++) {
	addObjectToSelection(objects_[i]) ;
    }
    updateGL() ;
}

void Viewer::unselectAll(void) 
{
    for(int i=0 ; i < selection_.size() ; i++)
	selection_[i]->deselect() ;
    selection_.clear();
    updateGL();
}

void Viewer::setMeshEditionMode(MeshEditionMode mode) 
{
    meshEditionMode_ = mode ;
    QList<QAxelObject *> l = listObjectsFromType("Mesh") ;
    for(int i=0 ; i < l.size() ; i++) {
	if(meshEditionMode_ == VertexEdition) ((QMesh *)l[i])->setEditionMode(QMesh::Vertex) ;
	if(meshEditionMode_ == EdgeEdition) ((QMesh *)l[i])->setEditionMode(QMesh::Edge) ;
	if(meshEditionMode_ == FaceEdition) ((QMesh *)l[i])->setEditionMode(QMesh::Face) ;
    }
}

void Viewer::formatInstructions(int width, int height)
{
    QString text = tr("Click and drag with the left mouse button "
                      "to rotate the Qt logo.");

    QFontMetrics metrics = QFontMetrics(font());
    int border = qMax(4, metrics.leading());

    QRect rect = metrics.boundingRect(0, 0, width - 2*border, int(height*0.125), Qt::AlignCenter | Qt::TextWordWrap, text);
    image = QImage(width, rect.height() + 2*border, QImage::Format_ARGB32_Premultiplied);
    image.fill(qRgba(0, 0, 0, 127));

    QPainter painter;
    painter.begin(&image);
    painter.setRenderHint(QPainter::TextAntialiasing);
    painter.setPen(Qt::white);
    painter.drawText((width - rect.width())/2, border,
                     rect.width(), rect.height(),
                     Qt::AlignCenter | Qt::TextWordWrap, text);
    painter.end();
}

void Viewer::showObject(QAxelObject * object, int msec)
{
    temporary_ << object ;
    update() ;
    QTimer * timer = new QTimer(this) ;
    connect(timer, SIGNAL(timeout()), this, SLOT(removeTemporaryObjects())) ;
    timer->start(msec) ;
}

void Viewer::removeTemporaryObjects(void)
{
    if(!temporary_.empty())
	temporary_.pop_front() ;
    update() ;
}

void Viewer::setOrthographicDisplay(void)
{
    camera()->setType(qglviewer::Camera::ORTHOGRAPHIC) ;
    update() ;
    emit cameraModeChanged(QString(" Orthographic ")) ;
}

void Viewer::setPerspectiveDisplay()
{
    camera()->setType(qglviewer::Camera::PERSPECTIVE) ;
    update() ;
    emit cameraModeChanged(QString(" Perspective ")) ;
}

void Viewer::set2dMode(void)
{
    setCameraPosition(0.0,0.0,camera()->position().z) ;
    setCameraUpVector(0.0,1.0,0.0) ;
    setCameraViewDirection(0.0,0.0,-3.0) ;
   
    setAxisIsDrawn(false) ;
    setOrthographicDisplay() ;
    cameraConstraint->setRotationConstraintType(AxisPlaneConstraint::FORBIDDEN) ;
    cameraConstraint->setTranslationConstraintType(AxisPlaneConstraint::FREE) ;
    setCameraConstraint(cameraConstraint);
    
    update() ;
    emit dimensionModeChanged(" 2D  ");
}

void Viewer::set3dMode(void)
{
    setAxisIsDrawn(true) ;
    setPerspectiveDisplay() ;
    cameraConstraint->setRotationConstraintType(AxisPlaneConstraint::FREE) ;
    cameraConstraint->setTranslationConstraintType(AxisPlaneConstraint::FREE) ;
    setCameraConstraint(cameraConstraint);

    update() ;
    emit dimensionModeChanged(" 3D  ");
    
}

void Viewer::setCameraPosition(float x, float y, float z)
{
    camera()->setPosition(Vec(x, y, z)) ;
}

void Viewer::setCameraUpVector(float x, float y, float z)
{
    camera()->setUpVector(Vec(x, y, z)) ;
}

void Viewer::setCameraViewDirection(float x, float y, float z)
{
    camera()->setViewDirection(Vec(x, y, z)) ;
}
void Viewer::setCameraConstraint(CameraConstraint * constraint)
{
    camera()->frame()->setConstraint(constraint) ;
}
