/*****************************************************************************
 * A x e l
 *****************************************************************************
 * ViewerActions
 * 2006-11-08
 * Julien Wintz
 *****************************************************************************
 *               Copyright (C) 2006 INRIA Sophia-Antipolis
 *****************************************************************************
 * Comments :
 ****************************************************************************/

# include <QFile>
# include <QCursor>
# include <QAction>
# include <QMenu>
# include <QMessageBox>

# include <QGui/Viewer.h>

# include <QAxel/QVertex.h>
# include <QAxel/QEdge.h>
# include <QAxel/QFace.h>
# include <QAxel/QMesh.h>

# include <QAxel/QArrow.h>
# include <QAxel/QBoundingBox.h>
# include <QAxel/QPoint3f.h>
# include <QAxel/QPlane.h>
# include <QAxel/QCurve.h>
# include <QAxel/QSurface.h>

# include <QSynaps/QSynapsInterface.h>
# include <QCore/QCoreInterface.h>
# include <QSisl/QSislInterface.h>

# include <QAxel/PluginManager.h>

# include <QGui/ToolPropertiesDock.h>
# include <QGui/ToolPropertiesWidgets.h>

# include <QIntersection/Curves/QBezierClipper.h>
# include <QIntersection/Surfaces/QRationalSurfaceIntersection.h>

using namespace std ;
using namespace QAxel ;
using namespace QIntersection ;

void Viewer::refineImplicitCurves(void)
{
    bool switchback = false ;
    if(camera()->type() == Camera::PERSPECTIVE) {
	camera()->setType(Camera::ORTHOGRAPHIC) ;
	switchback = true ;
    }
    GLdouble coef[6][4];
    camera()->getFrustumPlanesCoefficients(coef) ;
    double xmin = coef[0][0] * coef[0][3] ; cout << "xmin = " << xmin << endl ;
    double xmax = coef[1][0] * coef[1][3] ; cout << "xmax = " << xmax << endl ;
    double ymin = coef[5][1] * coef[5][3] ; cout << "ymin = " << ymin << endl ;
    double ymax = coef[4][1] * coef[4][3] ; cout << "ymax = " << ymax << endl ;
    double zmin = coef[3][2] * coef[3][3] ; cout << "zmin = " << zmin << endl ;
    double zmax = coef[2][2] * coef[2][3] ; cout << "zmax = " << zmax << endl ;
    if(switchback)	camera()->setType(Camera::PERSPECTIVE) ;
    
    foreach(QAxelObject * o, listSelectedObjectsFromType("ImplicitCurve2D")) {
	if(QImplicitCurve2D * c_old = dynamic_cast<QImplicitCurve2D *>(o)) {
	    QImplicitCurve2D * c_new = new QImplicitCurve2D(c_old->equation(), new QBoundingBox(xmin, xmax, ymin, ymax), QImplicitCurve::Subdivision, QImplicitCurve::Relative) ;
	    addObject(c_new) ;
	}
    }
}

void Viewer::intersectImplicitCurves(void)
{
    QList<QAxelObject *> selectedObjects ;
    selectedObjects << listSelectedObjectsFromType("ImplicitCurve2D") ;
    
    if(selectedObjects.size() != 2) {
	QMessageBox::information(0, "Axel","Implicit Curves : Select ONLY two curves","&Ok",0) ;
	return ;
    }
    
    QImplicitCurve2D * c1 = dynamic_cast<QImplicitCurve2D *>(selectedObjects[0]) ;
    QImplicitCurve2D * c2 = dynamic_cast<QImplicitCurve2D *>(selectedObjects[1]) ;
    QBoundingBox *     bb = c1->boundingbox()->intersected(c2->boundingbox()) ;
    
    QSynapsInterface * iSynaps = dynamic_cast<QSynapsInterface *>(PluginManager::instance()->plugin("Synaps")) ;
    if(iSynaps) {
	QList<QPoint3f *> points = iSynaps->intersection(c1, c2, bb) ;
	addObject(bb) ;
	foreach(QPoint3f * p, points) addObject(p) ;
    } else {
	QMessageBox::information(0, "Axel","Implicit curves intersection: QSynaps plugin is not loaded","&Ok",0) ;
    }
}

void Viewer::intersectParametricSurfaces(void)
{
    QList<QAxelObject *> selectedObjects ;
    selectedObjects << listSelectedObjectsFromType("RationalSurface") ;
    selectedObjects << listSelectedObjectsFromType("BSplineSurface") ;
  
    if(selectedObjects.size() != 2) {
	cout << "size : " << selectedObjects.size() << endl ;
	QMessageBox::information(0, "Axel","Parametric Surfaces Intersection : Select ONLY two surfaces","&Ok",0) ;
	return ;
    }
  
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    if(iCore) {
	QShape * intersection = iCore->intersect(selectedObjects[0], selectedObjects[1]) ;
	intersection->setRGB(255, 0, 0) ;
	addObject(intersection) ;
    } else {
	QMessageBox::information(0, "Axel","QRationalCurve Computation : QCore plugin is not loaded","&Ok",0) ;
    }
}

void Viewer::intersectImplicitSurfaces(void)
{
    QList<QAxelObject *> selectedObjects ;
    selectedObjects << listSelectedObjectsFromType("ImplicitSurface") ;
    selectedObjects << listSelectedObjectsFromType("Plane") ;	

    if(selectedObjects.size() != 2) {
	QMessageBox::information(0, "Axel","Implicit Surfaces Intersection : Select ONLY two surfaces","&Ok",0) ;
	return ;
    }
	
    QStringList equations ;
    QBoundingBox * bb = 0 ;
    foreach(QAxelObject * o, selectedObjects) {
	if(QImplicitSurface * s = dynamic_cast<QImplicitSurface *>(o)) { equations << s->equation() ; bb = s->boundingbox() ; }
	if(QPlane * p = dynamic_cast<QPlane *>(o)) equations << p->equation() ;
    }

    cout << "Implicit surfaces intersection:" << endl ;
    cout << "1) " << equations[0].toAscii().constData() << endl ;
    cout << "2) " << equations[1].toAscii().constData() << endl ;

    QImplicitCurve3D * intersection = new QImplicitCurve3D(equations[0], equations[1], bb) ;
    intersection->setRGB(175, 34, 255) ;
    addObject(intersection) ;
}

void Viewer::autoIntersectParametricSurfaces(void)
{
    QList<QAxelObject *> l ;
    l << listSelectedObjectsFromType("RationalSurface") ;
    l << listSelectedObjectsFromType("BSplineSurface") ;
	
    QCoreInterface * iCore = dynamic_cast<QCoreInterface *>(PluginManager::instance()->plugin("Core")) ;
    if(iCore) foreach(QAxelObject * o, l) {
	    QShape * intersection = iCore->autointersect(o) ;
	    intersection->setRGB(255, 0, 0) ;
	    addObject(intersection) ;
	} else {
	QMessageBox::information(0, "Axel","Parametric Surfaces Auto-Intersection : Core plugin not loaded","&Ok",0) ;
    }
}

void Viewer::approximatePointsAsBSplineCurve()
{
    QList<QAxelObject *> l ; l << listSelectedObjectsFromType("Point3f") ;
    QList<QPoint3f *> points ; foreach(QAxelObject * o, l) points << qobject_cast<QPoint3f *>(o) ;

    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
    if(iSisl) {
	QBSplineCurve * approximation = iSisl->approximateBSplineCurve(points, 1, 0, 4, 3) ;
	approximation->compute() ;
	addObject(approximation) ;
    } else {
	QMessageBox::information(0, "Axel","Interpolation : Sisl plugin not loaded","&Ok",0) ;
    }
}

void Viewer::approximatePointsAsBSplineSurface(int iopen1, int iopen2, int order1, int order2, int dimension)
{
    QList<QAxelObject *> l ; l << listSelectedObjectsFromType("Point3f") ;
    QList<QPoint3f *> points ; foreach(QAxelObject * o, l) points << qobject_cast<QPoint3f *>(o) ;

    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
    if(iSisl) {
	QBSplineSurface * approximation = iSisl->approximateBSplineSurface(points, iopen1, iopen2, order1, order2, dimension) ;
	if(approximation) {
	    approximation->compute() ;
	    addObject(approximation) ;
	}
    
    } else {
	QMessageBox::information(0, "Axel","Interpolation : Sisl plugin not loaded","&Ok",0) ;
    }
}

void Viewer::joinBSplineCurves(void)
{
    QList<QAxelObject *> l ; l << listSelectedObjectsFromType("BSplineCurve") ;
    
    if(l.size() !=2) {
	QMessageBox::information(0, "Axel","Select Two B-Spline Curves","&Ok",0) ;
	return ;
    }

    QList<QBSplineCurve *> curves ; foreach(QAxelObject * o, l) curves << qobject_cast<QBSplineCurve *>(o) ;

    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
    
    if(iSisl) {
	QBSplineCurve * join = iSisl->join(curves.first(), curves.last()) ;
	join->compute() ;
	addObject(join) ;
    } else {
	QMessageBox::information(0, "Axel","Interpolation : Sisl plugin not loaded","&Ok",0) ;
    }
}

void Viewer::interpolatePointsAsBSpline(void)
{    
    QList<QAxelObject *> l ; l << listObjectsFromType("Point3f") ;
    QList<QPoint3f *> points ; foreach(QAxelObject * o, l) points << dynamic_cast<QPoint3f *>(o) ;

    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
    if(iSisl) {
	QBSplineCurve * interpolation = iSisl->interpolate(points) ;
	interpolation->compute() ;
	addObject(interpolation) ;
    } else {
	QMessageBox::information(0, "Axel","Interpolation : Sisl plugin not loaded","&Ok",0) ;
    }
}

void Viewer::intersectBSplineCurves(void) 
{
    QList<QAxelObject *> l ; l << listSelectedObjectsFromType("BSplineCurve") ;
    QList<QBSplineCurve *> curves ; foreach(QAxelObject * o, l) curves << dynamic_cast<QBSplineCurve *>(o) ;
	
    if(l.size() != 2) {
	QMessageBox::information(0, "Axel","BSpline Curves intersection : Select ONLY two curves","&Ok",0) ;
	return ;
    }
	
    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
    if(iSisl) {
	QList<QPoint3f *> points = iSisl->intersection(curves[0], curves[1]) ;
	foreach(QPoint3f * p, points) addObject(p) ;
    } else {
	QMessageBox::information(0, "Axel","Blend BSpline Curves : Sisl plugin not loaded","&Ok",0) ;
    }
}

void Viewer::offsetBSplineCurves(void)
{
    QList<QAxelObject *> lc ; lc << listSelectedObjectsFromType("BSplineCurve") ;
    QList<QBSplineCurve *> curves ; foreach(QAxelObject * o, lc) curves << dynamic_cast<QBSplineCurve *>(o) ;

    QList<QAxelObject *> lp ; lp << listSelectedObjectsFromType("Point3f") ;
    QList<QPoint3f *> points ; foreach(QAxelObject * o, lp) points << dynamic_cast<QPoint3f *>(o) ;
	
    if(curves.size() != 1 && points.size() != 1) {
	QMessageBox::information(0, "Axel","Offset BSpline Curve : Select a curve and a point","&Ok",0) ;
	return ;
    }
	
    QBSplineCurve * c = curves[0] ;
    QPoint3f * p = points[0] ;

    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
	
    QPoint3f * cl = iSisl->closestPoint(c, p) ;
    cl->setRGB(255,0,0) ;
    addObject(cl) ;
	
    addObject(new QArrow(cl, p)) ;

    if(iSisl) {
	Vec v(p->x()-cl->x(), p->y()-cl->y(), p->z()-cl->z()) ;
	QBSplineCurve * offset_curve = iSisl->offset(c, v, v.norm()) ;
	offset_curve->compute() ;
	addObject(offset_curve) ;
    } else {
	QMessageBox::information(0, "Axel","Parametric Surfaces Auto-Intersection : Core plugin not loaded","&Ok",0) ;
    }
}

void Viewer::blendBSplineCurves(void) 
{
    QList<QAxelObject *> l ; l << listSelectedObjectsFromType("BSplineCurve") ;
    QList<QBSplineCurve *> curves ; foreach(QAxelObject * o, l) curves << dynamic_cast<QBSplineCurve *>(o) ;
	
    if(l.size() != 2) {
	QMessageBox::information(0, "Axel","BSpline Curves blending : Select ONLY two curves","&Ok",0) ;
	return ;
    }
	
    QSislInterface * iSisl = dynamic_cast<QSislInterface *>(PluginManager::instance()->plugin("Sisl")) ;
    if(iSisl) {
	QBSplineCurve * blend_curve = iSisl->blend(curves[0], curves[1]) ;
	blend_curve->compute() ;
	addObject(blend_curve) ;
    } else {
	QMessageBox::information(0, "Axel","Blend BSpline Curves : Sisl plugin not loaded","&Ok",0) ;
    }
}
using namespace QAxel;
void Viewer::intersectRationalSurfaces(void)
{
    QList<QAxelObject *> l ; l << listSelectedObjectsFromType("RationalSurface") ;
    QList<QRationalSurface *> surfaces ; foreach(QAxelObject * o, l) surfaces << dynamic_cast<QRationalSurface *>(o) ;

	    if(l.size() != 2) {
			// QMessageBox::information(0, "Axel","Rational Surface Intersection : Select ONLY two surfaces","&Ok",0) ;

			// Testing...
				QRationalSurface * S1= new QRationalSurface("s^2t^2+st^3","s^2t+st^2","s^2t^2+st-t^2","1",0,1,0,1);
				QRationalSurface * S2 = new QRationalSurface("3u^2+u^2v^3","u^2v^2+uv^3","2u^2v^2-v^4","1",0,1,0,1);
				// addObject(S1);
				// addObject(S2);
				QRationalSurfaceIntersection * algo = new QRationalSurfaceIntersection(S1, S2, dynamic_cast<MainWindow *>(parent())) ;
				algo->run() ;
	
			// QPoint3f * p = new QPoint3f(0.0,0.0,0.0) ;


			// QImplicitCurve2D * C1 = new QImplicitCurve2D("x+y", new QBoundingBox(-10.001,10.1,-10.01,10.0001), QCurve::Subdivision);
			// addObject(C1);
			// End of testing
			
		return ;
	    } 
		else {
			QRationalSurfaceIntersection * algo = new QRationalSurfaceIntersection(surfaces[0], surfaces[1], dynamic_cast<MainWindow *>(parent())) ;
			algo->run() ;
			// QDiscreteCurve * c = algo->getCurve() ;
			// addObject(c) ;
    	}
}

void Viewer::quadClipImplicitCurves(void)
{
    MainWindow * mw = dynamic_cast<MainWindow *>(parent()) ;
    mw->toolPropertiesDock()->createBezierClipperWidget() ;
}
