/*****************************************************************************
 * A x e l
 *****************************************************************************
 * QBezierClipper
 * 2007-04-02
 * Julien Wintz
 * Brian Moore
 *****************************************************************************
 *               Copyright (C) 2007 INRIA Sophia-Antipolis
 *               Copyright (C) 2007 RICAM
 *****************************************************************************
 * Comments:
 * - There is only one box computed each time
 * - The error has to be REALLY computed
 * - The Quad subdivision of the cell also has to be implemented ( following 
 *   which criteria) ?
 *****************************************************************************/

# include <QGui/Viewer.h>

# include <QAxel/PluginManager.h>

# include <QIntersection/Curves/QBezierClipper.h>
# include <QSynaps/QSynapsInterface.h>

using namespace std ;

QBezierClipper::QBezierClipper(QImplicitCurve2D * c1, QImplicitCurve2D * c2, MainWindow * mw)
{
    this->c_c1 = c1 ;
    this->c_c2 = c2 ;
    this->c_mw = mw ;
    
    this->cell = new QCell(c_c1->boundingbox()->intersect(c2->boundingbox())) ;

    this->c_step = 0 ;
    this->c_epsilon = 0.1 ;
    this->c_showLinearGraphedSurfaces = true ;
    this->c_showLinearOffsetSurfaces = false ;
    this->c_showQuadraticGraphedSurfaces = true ;
    this->c_showQuadraticOffsetSurfaces = false ;
    this->c_showAllCriticalPoints = false ;
    this->c_showOnlyRelevantCriticalPoints = false ;

    this->c_linearGraphedSurfacesIsDisplayed = false;
    this->c_quadraticGraphedSurfacesIsDisplayed = false;

    this->c_linearOffsetSurfacesIsDisplayed = false;
    this->c_quadraticOffsetSurfacesIsDisplayed = false;

}

QImplicitSurface * QBezierClipper::graph(QImplicitCurve2D * c)
{
    QString equation = c->equation() ;
    equation.append("-z") ;

    QBoundingBox * bb = new QBoundingBox(
	cell->xmin(),
	cell->xmax(),
	cell->ymin(),
	cell->ymax(),
	-5.0, 5.0
    ) ;
    bb->setXMin(bb->xmin()-0.01) ;
    bb->setXMax(bb->xmax()+0.02) ;
    bb->setYMin(bb->ymin()-0.01) ;
    bb->setYMax(bb->ymax()+0.02) ;
    bb->setZMin(bb->zmin()-0.01) ;
    bb->setZMax(bb->zmax()+0.02) ;

    QImplicitSurface * s = new QImplicitSurface(equation, bb) ;
    return s ;
}

void QBezierClipper::linearApproximation(void)
{
    if(c_showLinearGraphedSurfaces) 
	if ( c_linearGraphedSurfacesIsDisplayed ) 
	    c_mw->viewer()->removeObject(c_s2) ;
	else
	    c_linearGraphedSurfacesIsDisplayed = true;

    c_s1 = graph(c_c1) ; c_s1->setSingularitiesDrawn(false) ; c_s1->setName("p_graph") ;
    if(c_showLinearGraphedSurfaces) c_mw->viewer()->addObject(c_s1) ;

    c_c1->setBBDrawn(true) ;

    QSynapsInterface * iSynaps = dynamic_cast<QSynapsInterface *>(PluginManager::instance()->plugin("Synaps")) ;
    int deg_x = iSynaps->degree(c_c1,0) ;
    int deg_y = iSynaps->degree(c_c1,1) ; 
    float * array = new float[(deg_x+1)*(deg_y+1)] ;
    iSynaps->toBezierArray(c_c1, array, &deg_x, &deg_y) ;
    c_linearApproximation = mult(array, deg_x, deg_y, 4, (deg_x+1)*(deg_y+1)) ;

//     for(int i = 0 ; i < 4 ; i++) cout << "L[" << i << "]=" << c_linearApproximation[i] << std::endl ;

    computeLinearErrorBound(array, c_linearApproximation, deg_x, deg_y) ;

    QString eq = iSynaps->toLinearMonomialEquation(c_linearApproximation, 
					     cell->xmin(),
					     cell->xmax(),
					     cell->ymin(),
					     cell->ymax()) ;
    QString error ; error.setNum(c_linearErrorBound) ; 
    
    cout << "about to compute " << eq.toAscii().constData() << endl ;

    c_p  = new QImplicitCurve2D(eq              , cell, QImplicitCurve::None) ; c_p->setName("linapp") ;
    c_p1 = new QImplicitCurve2D(eq + "-" + error, cell, QImplicitCurve::None) ; c_p1->setName("linapp1") ;
    c_p2 = new QImplicitCurve2D(eq + "+" + error, cell, QImplicitCurve::None) ; c_p1->setName("linapp2") ;

    if(c_showLinearOffsetSurfaces)
	if ( c_linearOffsetSurfacesIsDisplayed ) {
	    c_mw->viewer()->removeObject(s_p1) ;
	    c_mw->viewer()->removeObject(s_p2) ;
	} else {
	    c_linearOffsetSurfacesIsDisplayed = true;
	}

    s_p1 = graph(c_p1) ; s_p1->setName("s_p1") ; s_p1->setRGBA(110, 139, 61, 220) ; 
    s_p2 = graph(c_p2) ; s_p2->setName("s_p2") ; s_p2->setRGBA(110, 139, 61, 220) ; 

    if(c_showLinearOffsetSurfaces) {
	c_mw->viewer()->addObject(s_p1) ;
	c_mw->viewer()->addObject(s_p2) ;
    }
}

/*! Short description.
 *  Long description.
 */
void QBezierClipper::quadraticApproximation(void)
{
    if(c_showQuadraticGraphedSurfaces) 
    {
	if ( c_quadraticGraphedSurfacesIsDisplayed ) 
	{
	    c_mw->viewer()->removeObject(c_s2) ;
	}
	else
	{
	    c_quadraticGraphedSurfacesIsDisplayed = true;
	}
    }

    c_s2 = graph(c_c2) ; c_s2->setSingularitiesDrawn(false) ; c_s2->setName("q_graph") ;
    if(c_showQuadraticGraphedSurfaces) c_mw->viewer()->addObject(c_s2) ;

    QSynapsInterface * iSynaps = dynamic_cast<QSynapsInterface *>(PluginManager::instance()->plugin("Synaps")) ;

    int deg_x = iSynaps->degree(c_c2,0) ;
    int deg_y = iSynaps->degree(c_c2,1) ; 
    float * array = new float[(deg_x+1)*(deg_y+1)] ;

    iSynaps->toBezierArray(c_c2, array, &deg_x, &deg_y) ;
    c_quadraticApproximation = mult(array, deg_x, deg_y, 9, (deg_x+1)*(deg_y+1)) ;
    computeQuadraticErrorBound(array, c_quadraticApproximation, deg_x, deg_y) ;

    // for(int i = 0 ; i < 9 ; i++) cout << "K[" << i << "]=" << c_quadraticApproximation[i] << std::endl ;

    QString eq = iSynaps->toQuadraticMonomialEquation(c_quadraticApproximation, 
					     cell->xmin(),
					     cell->xmax(),
					     cell->ymin(),
					     cell->ymax()) ;

    QString error ; error.setNum(c_quadraticErrorBound) ;
    c_q  = new QImplicitCurve2D(eq              , cell, QImplicitCurve::None) ; c_q->setName("quadapp") ;
    c_q1 = new QImplicitCurve2D(eq + "-" + error, cell, QImplicitCurve::None) ; c_q1->setName("quadapp1") ;
    c_q2 = new QImplicitCurve2D(eq + "+" + error, cell, QImplicitCurve::None) ; c_q2->setName("quadapp2") ;

    if(c_showQuadraticOffsetSurfaces) {
	if ( c_quadraticOffsetSurfacesIsDisplayed ) 
	{
	    c_mw->viewer()->removeObject(s_q1) ;
	    c_mw->viewer()->removeObject(s_q2) ;
	}
	else
	{
	    c_quadraticOffsetSurfacesIsDisplayed = true;
	}

    }

    s_q1 = graph(c_q1) ; s_q1->setName("s_q1") ; s_q1->setRGBA(110, 139, 61, 220) ; 
    s_q2 = graph(c_q2) ; s_q2->setName("s_q2") ; s_q2->setRGBA(110, 139, 61, 220) ; 

    if(c_showQuadraticOffsetSurfaces) {
	c_mw->viewer()->addObject(s_q1) ;
	c_mw->viewer()->addObject(s_q2) ;
    }
}




/* This function returns the kth control point(the z-value) of a curve of degree "degree"
   under degree elevation to "elevationDegree" */
 
float QBezierClipper::computeCoeff(float * curve, int k, int degree, int elevationDegree) {
    float s;
    int binomial_mk;
    float temp;
    int degreeDiff;

    /* Initialization */
    s = 0;

    degreeDiff = elevationDegree - degree;

    QSynapsInterface * iSynaps = dynamic_cast<QSynapsInterface *>(PluginManager::instance()->plugin("Synaps")) ;

    binomial_mk = iSynaps->binomial(elevationDegree, k);
    temp = (float)binomial_mk;
//     cout << "  Binomial (" << elevationDegree << " , " << k << ") = " << binomial_mk << ", denominator = " << temp << endl;
    for(int i=0; i <= degree ; i++) {
	if(k-i >= 0 && degreeDiff >= k-i) { 
	    s = s + curve[i]*(iSynaps->binomial(degree,i))*(iSynaps->binomial(degreeDiff, k-i))/temp;
//         	cout << "  i = " << i << ", s = " << s << ",  " << curve[i] << ", " << iSynaps->binomial(degree,i) << ", " << iSynaps->binomial(degreeDiff, k-i) << ", " << degreeDiff << ", " << k-i <<endl; 
	}
    }
    return s;
}





float * QBezierClipper::degreeElevationSurface(float *initSurface , int currentDegreeY, int currentDegreeX, int degreeY, int degreeX) {
    float * interSurface, * newSurface, * col;
    int newNumRows, newNumColumns, nRows, nColumns;

    nRows = currentDegreeY + 1;
    nColumns =  currentDegreeX + 1;

    newNumRows = degreeY + 1;
    newNumColumns = degreeX + 1;

    interSurface = new float[newNumRows * newNumColumns];
    newSurface = new float[newNumRows * newNumColumns];

    col = new float[nRows];

    /* Validation of the inputs */
    if((newNumRows < nRows) || (newNumColumns < nColumns)) {
	cout << "ERROR: Invalid degree for degree elevation " << endl;
    }

    /* Elevation in x */
    for(int i=0; i < nRows; i++) {
	for(int j=0; j < newNumColumns; j++) {
	    interSurface[i*newNumColumns + j] = computeCoeff(&initSurface[i*nColumns], j, nColumns-1, newNumColumns-1);
	}
    }


    /* Elevation in y */
    for(int j=0; j < newNumColumns; j++) {

	/* Extract column j */
	for(int k=0; k<nRows; k++) {
	    col[k] = interSurface[k*newNumColumns + j];
	}

	for(int i=0; i < newNumRows; i++) {
	    newSurface[i*newNumColumns + j] = computeCoeff(col, i, nRows-1, newNumRows-1);
	}
    }

    return newSurface;
}












void QBezierClipper::computeLinearErrorBound(float * array, float * approximation, int deg_x, int deg_y)
{

    float maxError = 0.0, z1, z2, dz, err, zLine, ratio;

    for(int i=0; i <= deg_x; i++) {
	/* Find the two z points on the horizontal axis i */
	z1 = approximation[0] + (approximation[2]-approximation[0])*(i/(float)(deg_x));
	z2 = approximation[1] + (approximation[3]-approximation[1])*(i/(float)(deg_x));
	dz = z2 - z1;
	
	for(int j=0; j <= deg_y; j++) {
	    /* Compute the z value of the control point at position (i,j) */
	    ratio = j/(float)(deg_y);
	    zLine = z1 + ratio*dz;
	    
	    err = fabs(array[i*(deg_y+1)+j] - zLine);
	    if(err > maxError) {
		maxError = err;
	    }
	}
    }

    c_linearErrorBound = maxError;
}




void QBezierClipper::computeQuadraticErrorBound(float * array, float * approximation, int deg_x, int deg_y)
{
    
    int nColumns, nRows;
    float * elevatedConic, err, maxError;

    nRows = deg_x + 1;
    nColumns = deg_y + 1;

    if(nRows < 3 || nColumns < 3) {
	cout << "ERROR: THe degree is too small !!!" << endl;
    }

    /* Compute degree elevation */
    elevatedConic = degreeElevationSurface(approximation, 2, 2, deg_x, deg_y);

    maxError = 0;
    for(int i=0; i<nRows*nColumns; i++) {
	err = fabs(array[i]-elevatedConic[i]);
	if(err > maxError) {
	    maxError = err;
	} 	
    }
    c_quadraticErrorBound = maxError;
}







void QBezierClipper::clip(void)
{
    cout << __func__ << endl ;

    QSynapsInterface * iSynaps = dynamic_cast<QSynapsInterface *>(PluginManager::instance()->plugin("Synaps")) ;

    QList<QPoint3f *> intersectionsP1Q1 = iSynaps->intersection(c_p1, c_q1, cell) ;
    QList<QPoint3f *> intersectionsP1Q2 = iSynaps->intersection(c_p1, c_q2, cell) ;
    QList<QPoint3f *> intersectionsP2Q1 = iSynaps->intersection(c_p2, c_q1, cell) ;
    QList<QPoint3f *> intersectionsP2Q2 = iSynaps->intersection(c_p2, c_q2, cell) ;

    QList<QPoint3f *> intersectionsP1BB = iSynaps->intersectionPoints(c_p1, cell) ;
    QList<QPoint3f *> intersectionsP2BB = iSynaps->intersectionPoints(c_p2, cell) ;
    QList<QPoint3f *> intersectionsQ1BB = iSynaps->intersectionPoints(c_q1, cell) ;
    QList<QPoint3f *> intersectionsQ2BB = iSynaps->intersectionPoints(c_q2, cell) ;

    QList<QPoint3f *> xtremPoints ;
    xtremPoints << iSynaps->xtremPoints(c_q1, cell) ;
    xtremPoints << iSynaps->xtremPoints(c_q2, cell) ;

    QList<QPoint3f *> ytremPoints ;
    ytremPoints << iSynaps->ytremPoints(c_q1, cell) ;
    ytremPoints << iSynaps->ytremPoints(c_q2, cell) ;

    QList<QPoint3f *> corners = c_c1->boundingbox()->corners() ;

    foreach(QPoint3f * p, intersectionsP1Q1) { p->setSize(0.1) ; p->setColor(Qt::green) ; p->blacklist() ; } 
    foreach(QPoint3f * p, intersectionsP1Q2) { p->setSize(0.1) ; p->setColor(Qt::green) ; p->blacklist() ; } 
    foreach(QPoint3f * p, intersectionsP2Q1) { p->setSize(0.1) ; p->setColor(Qt::green) ; p->blacklist() ; } 
    foreach(QPoint3f * p, intersectionsP2Q2) { p->setSize(0.1) ; p->setColor(Qt::green) ; p->blacklist() ; } 

    foreach(QPoint3f * p, intersectionsP1BB) { p->setSize(0.1) ; p->setColor(Qt::red) ; p->blacklist() ; } 
    foreach(QPoint3f * p, intersectionsP2BB) { p->setSize(0.1) ; p->setColor(Qt::red) ; p->blacklist() ; } 
    foreach(QPoint3f * p, intersectionsQ1BB) { p->setSize(0.1) ; p->setColor(Qt::red) ; p->blacklist() ; } 
    foreach(QPoint3f * p, intersectionsQ2BB) { p->setSize(0.1) ; p->setColor(Qt::red) ; p->blacklist() ; } 

    foreach(QPoint3f * p, xtremPoints) { p->setSize(0.1) ; p->setColor(Qt::blue) ; p->blacklist() ; } 
    foreach(QPoint3f * p, ytremPoints) { p->setSize(0.1) ; p->setColor(Qt::blue) ; p->blacklist() ; } 

    foreach(QPoint3f * p, corners) { p->setSize(0.1) ; p->setColor(Qt::yellow) ; p->blacklist() ; } 

    /* A. We keep the following points:
     * 1) Pt Intersection: Line-Box                  -> check if between conics
     * 2) Pt Intersection: Conic-Box                 -> check if between lines
     * 3) Pt Intersection: Box corners               -> check if between lines and if between conics
     * 4) Pt Intersection: Tangent (critical points) -> check if between lines
     * 5) Pt Intersection: Line-conics               -> keep all
     */
    
    QList<QPoint3f *> finalPoints ;
    
    foreach(QPoint3f * p, intersectionsP1BB) if(iSynaps->isOn(p, c_q, c_quadraticErrorBound)) finalPoints << p ;
    foreach(QPoint3f * p, intersectionsP2BB) if(iSynaps->isOn(p, c_q, c_quadraticErrorBound)) finalPoints << p ;

    foreach(QPoint3f * p, intersectionsQ1BB) if(iSynaps->isOn(p, c_p, c_linearErrorBound)) finalPoints << p ;
    foreach(QPoint3f * p, intersectionsQ2BB) if(iSynaps->isOn(p, c_p, c_linearErrorBound)) finalPoints << p ;

    foreach(QPoint3f * p, corners) if(iSynaps->isOn(p, c_p, c_linearErrorBound) && iSynaps->isOn(p, c_q, c_quadraticErrorBound)) finalPoints << p ;

    foreach(QPoint3f * p, xtremPoints) if(iSynaps->isOn(p, c_p, c_linearErrorBound)) finalPoints << p ;
    foreach(QPoint3f * p, ytremPoints) if(iSynaps->isOn(p, c_p, c_linearErrorBound)) finalPoints << p ;

    foreach(QPoint3f * p, intersectionsP1Q1) finalPoints << p ;
    foreach(QPoint3f * p, intersectionsP1Q2) finalPoints << p ;
    foreach(QPoint3f * p, intersectionsP2Q1) finalPoints << p ;
    foreach(QPoint3f * p, intersectionsP2Q2) finalPoints << p ;

    if(c_showAllCriticalPoints) {
	if(c_showOnlyRelevantCriticalPoints) {
	    foreach(QPoint3f * p, finalPoints) c_mw->viewer()->addObject(p) ;
	} else {
	    foreach(QPoint3f * p, intersectionsP1Q1) c_mw->viewer()->addObject(p) ;
	    foreach(QPoint3f * p, intersectionsP1Q2) c_mw->viewer()->addObject(p) ;
	    foreach(QPoint3f * p, intersectionsP2Q1) c_mw->viewer()->addObject(p) ;
	    foreach(QPoint3f * p, intersectionsP2Q2) c_mw->viewer()->addObject(p) ;
	    
	    foreach(QPoint3f * p, intersectionsP1BB) c_mw->viewer()->addObject(p) ;
	    foreach(QPoint3f * p, intersectionsP2BB) c_mw->viewer()->addObject(p) ;
	    foreach(QPoint3f * p, intersectionsQ1BB) c_mw->viewer()->addObject(p) ;
	    foreach(QPoint3f * p, intersectionsQ2BB) c_mw->viewer()->addObject(p) ;

	    foreach(QPoint3f * p, xtremPoints) c_mw->viewer()->addObject(p) ;
	    foreach(QPoint3f * p, ytremPoints) c_mw->viewer()->addObject(p) ;

	    foreach(QPoint3f * p, corners) c_mw->viewer()->addObject(p) ;
	}
    }

    /* Compute the boxes.
     * Currently here is only one box computed from the list of relevant points.
     */

    float xmin = cell->center()->x(), xmax = cell->center()->x(), ymin = cell->center()->y(), ymax = cell->center()->y() ;

    if(finalPoints.size()) {
	xmin = finalPoints.first()->x() ; foreach(QPoint3f * p, finalPoints) if(p->x() < xmin) xmin = p->x() ;
	xmax = finalPoints.first()->x() ; foreach(QPoint3f * p, finalPoints) if(p->x() > xmax) xmax = p->x() ;
	ymin = finalPoints.first()->y() ; foreach(QPoint3f * p, finalPoints) if(p->y() < ymin) ymin = p->y() ;
	ymax = finalPoints.first()->y() ; foreach(QPoint3f * p, finalPoints) if(p->y() > ymax) ymax = p->y() ;
    }
    
    QBoundingBox * bb = new QBoundingBox(xmin, xmax, ymin, ymax) ; c_mw->viewer()->addObject(bb) ;
    cell = new QCell(bb) ;

    this->c_step++ ;
}

void QBezierClipper::run(void)
{
    this->c_showLinearGraphedSurfaces = false ;
    this->c_showLinearOffsetSurfaces = false ;
    this->c_showQuadraticGraphedSurfaces = false ;
    this->c_showQuadraticOffsetSurfaces = false ;
    this->c_showAllCriticalPoints = false ;
    this->c_showOnlyRelevantCriticalPoints = false ;

    while(cell->size() > c_epsilon) {
	linearApproximation() ;
	quadraticApproximation() ;
	clip() ;
    }

    QPoint3f * intersection = cell->center() ; intersection->setColor(Qt::red) ; c_mw->viewer()->addObject(intersection) ;
    cout << "done." << endl ;
}


